import re
from typing import List


class JsonlPreProcessor:

    def __init__(self):
        pass

    def split_jsonl_objects(self, original_content: str) -> List[str]:
        pattern_to_find_beginning_of_objects = re.compile(r'^(?!\n)\s*(?=\{)', re.RegexFlag.MULTILINE)
        parts = pattern_to_find_beginning_of_objects.split(original_content)

        result = []
        for part in parts:
            part = part.strip()

            # edge case for first match
            if "}" not in part:
                continue

            # ignore everything after the last closing bracket
            part = part.rsplit("}", maxsplit=1)[0] + "}"

            result.append(part)

        return result

    def cleanup_object_for_json_parsing(self, content: str) -> str:
        """
        Prepares the given content for json parsing (if possible)
        :param content: original content
        :return: cleaned up content
        """
        result = self._remove_comments(content).strip()
        result = self._remove_trailing_comma_of_last_property(result)
        return result

    def _remove_comments(self, content: str) -> str:
        """
        Removes comments indicated by "//"
        :param content: original content
        :return: modified content without comments
        """
        result_lines = []
        for line in content.splitlines():
            quote_count = 0
            new_line = []
            i = 0
            while i < len(line):
                char = line[i]
                if char == '"':
                    quote_count += 1
                if char == '/' and i + 1 < len(line) and line[i + 1] == '/' and quote_count % 2 == 0:
                    break
                new_line.append(char)
                i += 1
            cleaned_line = "".join(new_line).strip()
            if cleaned_line:  # Only append non-empty lines
                result_lines.append(cleaned_line)
        result = "\n".join(result_lines)
        return result.strip()

    def _remove_trailing_comma_of_last_property(self, content: str) -> str:
        """
        Removes the "," of the last property within the object.
        :param content: original content
        :return: modified content
        """
        result = content
        matches = re.findall(r",\s*}", content, flags=re.MULTILINE)
        if matches is not None:
            for match in matches:
                result = result.replace(match, match[1:])

        return result



import asyncio
import json
import uuid
from typing import Callable, List, Dict

from aiomqtt import Client, Message


class MqttClient:

    def __init__(self, host: str, port: int, mqtt_user: str, mqtt_password: str):
        self._mqtt_client_id = 'openhasp-config-manager'
        self._host = host
        self._port = port
        self._mqtt_user = mqtt_user
        self._mqtt_password = mqtt_password
        self._reconnect_interval_seconds = 5

        self._mqtt_client_task: asyncio.Task | None = None
        self._callbacks: Dict[str, List[callable]] = {}
        self.__mqtt_client: Client | None = None

    async def publish(self, topic: str, payload: any):
        """
        Publish a message to a topic
        :param topic: topic to publish to
        :param payload: payload to publish
        """
        async with self._create_mqtt_client() as client:
            if isinstance(payload, dict) or isinstance(payload, list):
                payload = json.dumps(payload)

            await client.publish(topic, payload=payload)

    async def subscribe(self, topic: str, callback: Callable):
        """
        Subscribe to a topic and call the callback when a message is received
        :param topic: topic to subscribe to
        :param callback: function to call when a message is received
        """
        if topic not in self._callbacks:
            self._callbacks[topic] = []
        if callback not in self._callbacks[topic]:
            self._callbacks[topic].append(callback)

        if self._mqtt_client_task is None:
            await self._start_mqtt_client_task()

    async def cancel_callback(self, topic: str = None, callback: Callable = None):
        """
        Cancel a single subscription callback or all callbacks for a specific topic
        :param topic: the topic to cancel all callbacks for
        :param callback: the specific callback to cancel
        """
        if topic is None and callback is None:
            raise ValueError('Must specify either topic, callback or both')

        if topic is not None and callback is not None:
            self._callbacks[topic] = list(filter(lambda x: x != callback, self._callbacks[topic]))

        elif topic is not None and topic in self._callbacks:
            self._callbacks.pop(topic)
        else:
            for topic in self._callbacks.keys():
                self._callbacks[topic] = list(filter(lambda x: x != callback, self._callbacks[topic]))

        if not any(self._callbacks.values()):
            await self._stop_mqtt_client_task()

    def _create_mqtt_client(self) -> Client:
        return Client(
            hostname=self._host,
            port=self._port,
            username=self._mqtt_user,
            password=self._mqtt_password,
            identifier=f"{self._mqtt_client_id}-{uuid.uuid4()}",
        )

    async def _start_mqtt_client_task(self):
        self._mqtt_client_task = asyncio.ensure_future(self._mqtt_client_task_function())

    async def _stop_mqtt_client_task(self):
        if self._mqtt_client_task is not None:
            self._mqtt_client_task.cancel()
            self._mqtt_client_task = None

    async def _mqtt_client_task_function(self):
        while True:
            try:
                async with self._create_mqtt_client() as client:
                    await client.subscribe("hasp/#")
                    async for message in client.messages:
                        await self._handle_message(message)
            except asyncio.CancelledError:
                break
            except Exception as ex:
                # TODO: use logger instead of print
                print(f'Error: {ex}; Reconnecting in {self._reconnect_interval_seconds} seconds ...')
                await asyncio.sleep(self._reconnect_interval_seconds)

    async def _handle_message(self, message: Message):
        for topic, callbacks in self._callbacks.items():
            if message.topic.matches(topic):
                for callback in callbacks:
                    await callback(message.topic, message.payload)



import asyncio
import os
import re
import sys

from telnetlib3 import TelnetTerminalClient, open_connection, telnet_client_shell, TelnetWriterUnicode


class OpenHaspTelnetClient:

    def __init__(self, host: str, port: int = 23, baudrate: int = 115200, user: str = 'admin', password: str = 'admin'):
        self._host = host
        self._port = port
        self._baudrate = baudrate
        self._username = user
        self._password = password

    async def shell(self):
        async def _shell(reader, writer):
            while True:
                outp = await reader.read(1024)

                if not outp:
                    # End of File
                    break

                if 'Username:' in outp:
                    # reply all questions with 'y'.
                    writer.write(f"{self._username}\n")
                elif 'Password:' in outp:
                    writer.write(f"{self._password}\n")
                    break

            # switch over to interactive shell after automatic login
            await telnet_client_shell(reader, writer)

        reader, writer = await open_connection(
            host=self._host, port=self._port,
            tspeed=(self._baudrate, self._baudrate),
            shell=_shell,
            client_factory=TelnetTerminalClient
        )

        await writer.protocol.waiter_closed

    async def logs(self):
        async def _shell(reader, writer):
            login_done = False
            stdin, stdout = await self._make_stdio()
            buffer = ""
            while True:
                buffer = buffer + await reader.read(2048)

                if not buffer:
                    raise EOFError("Connection closed by remote host")

                if login_done:
                    buffer = buffer.replace('Prompt >', '')
                    buffer = re.sub(r"\x1b\[\d+\w+\s*", '', buffer, flags=re.IGNORECASE | re.S)
                    if not buffer.endswith("\r\n"):
                        # wait for the end of the line before processing
                        continue

                    lines = buffer.splitlines(keepends=True)
                    for line in lines:
                        line = line.lstrip(" ")
                        if not line.startswith("["):
                            continue
                        stdout.write(line.encode() or b":?!?:")

                    buffer = ""
                    continue

                if 'Username:' in buffer:
                    # reply all questions with 'y'.
                    writer.write(f"{self._username}\n")
                elif 'Password:' in buffer:
                    writer.write(f"{self._password}\n")
                    login_done = True

                buffer = ""
            # switch over to interactive shell after automatic login
            # await telnet_client_shell(reader, SilentTelnetWriterUnicode(writer, client=True))

        reader, writer = await open_connection(
            host=self._host, port=self._port,
            tspeed=(self._baudrate, self._baudrate),
            shell=_shell,
            client_factory=TelnetTerminalClient
        )

        await writer.protocol.waiter_closed

    async def _make_stdio(self):
        """
        Return (reader, writer) pair for sys.stdin, sys.stdout.

        This method is a coroutine.
        """
        reader = asyncio.StreamReader()
        reader_protocol = asyncio.StreamReaderProtocol(reader)

        # Thanks:
        #
        #   https://gist.github.com/nathan-hoad/8966377
        #
        # After some experimentation, this 'sameopenfile' conditional seems
        # allow us to handle stdin as a pipe or a keyboard.  In the case of
        # a tty, 0 and 1 are the same open file, we use:
        #
        #    https://github.com/orochimarufan/.files/blob/master/bin/mpr
        write_fobj = sys.stdout
        if os.path.sameopenfile(0, 1):
            write_fobj = sys.stdin
        loop = asyncio.get_event_loop_policy().get_event_loop()
        writer_transport, writer_protocol = await loop.connect_write_pipe(
            asyncio.streams.FlowControlMixin, write_fobj
        )

        writer = asyncio.StreamWriter(writer_transport, writer_protocol, None, loop)

        await loop.connect_read_pipe(lambda: reader_protocol, sys.stdin)

        return reader, writer


class SilentTelnetWriterUnicode(TelnetWriterUnicode):

    def __init__(self, writer, **kwargs):
        super().__init__(
            writer.transport, writer.protocol, writer.fn_encoding, encoding_errors=writer.encoding_errors,
            **kwargs
        )

    def _handle_do_forwardmask(self, buf):
        pass

    def write(self, string, errors=None):
        pass

    def writelines(self, lines, errors=None):
        pass

    def echo(self, string, errors=None):
        pass



from pathlib import Path

from openhasp_config_manager.cli.common import _create_config_manager, _analyze_and_filter
from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
from openhasp_config_manager.ui.util import success, error, info


async def c_shell(config_dir: Path, device: str):
    try:
        config_manager = _create_config_manager(config_dir, Path("./nonexistent"))
        filtered_devices, ignored_devices = _analyze_and_filter(config_manager=config_manager, device_filter=device)
        if len(filtered_devices) <= 0:
            raise Exception(f"No device matches the filter: {device}")
        if len(filtered_devices) > 1:
            raise Exception(f"More than one device matches the filter: {device}")
        info(f"Opening shell for: {filtered_devices[0].name}")
        for device in filtered_devices:
            client = OpenHaspClient(device)
            await client.shell()

        success("Done!")
    except Exception as ex:
        error(str(ex))
        raise ex



import asyncio
from enum import Enum
from typing import Dict, List, Any, Callable, Tuple

from aiomqtt import Topic

from openhasp_config_manager.openhasp_client.image_processor import OpenHaspImageProcessor
from openhasp_config_manager.openhasp_client.model.configuration.gui_config import GuiConfig
from openhasp_config_manager.openhasp_client.model.configuration.hasp_config import HaspConfig
from openhasp_config_manager.openhasp_client.model.configuration.http_config import HttpConfig
from openhasp_config_manager.openhasp_client.model.configuration.mqtt_config import MqttConfig
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.openhasp_client.mqtt_client import MqttClient
from openhasp_config_manager.openhasp_client.telnet_client import OpenHaspTelnetClient
from openhasp_config_manager.openhasp_client.webservice_client import WebserviceClient
from openhasp_config_manager.ui.util import info


class IntegratedIcon(Enum):
    """
    Definition of the integrated icons used in OpenHASP.
    """
    ARROW_DOWN = ("\uE045", "arrow-down")
    BATTERY_HIGH = ("\uF2A3", "battery-high")
    ARROW_DOWN_BOX = ("\uE6C0", "arrow-down-box")
    BATTERY_LOW = ("\uF2A1", "battery-low")
    ARROW_LEFT = ("\uE04D", "arrow-left")
    BATTERY_MEDIUM = ("\uF2A2", "battery-medium")
    ARROW_RIGHT = ("\uE054", "arrow-right")
    BATTERY_OUTLINE = ("\uE08E", "battery-outline")
    ARROW_UP = ("\uE05D", "arrow-up")
    EV_STATION = ("\uE5F1", "ev-station")
    ARROW_UP_BOX = ("\uE6C3", "arrow-up-box")
    LEAF = ("\uE32A", "leaf")
    CHEVRON_DOWN = ("\uE140", "chevron-down")
    LIGHTNING_BOLT = ("\uF40B", "lightning-bolt")
    CHEVRON_LEFT = ("\uE141", "chevron-left")
    POWER = ("\uE425", "power")
    CHEVRON_RIGHT = ("\uE142", "chevron-right")
    POWER_PLUG = ("\uE6A5", "power-plug")
    CHEVRON_UP = ("\uE143", "chevron-up")
    CEILING_LIGHT = ("\uE769", "ceiling-light")
    SUBDIRECTORY_ARROW_LEFT = ("\uE60C", "subdirectory-arrow-left")
    COACH_LAMP = ("\uF020", "coach-lamp")
    AIR_CONDITIONER = ("\uE01B", "air-conditioner")
    DESK_LAMP = ("\uE95F", "desk-lamp")
    CLOUD_SEARCH_OUTLINE = ("\uE957", "cloud-search-outline")
    FLOOR_LAMP = ("\uE8DD", "floor-lamp")
    FIRE = ("\uE238", "fire")
    LAMP = ("\uE6B5", "lamp")
    RADIATOR = ("\uE438", "radiator")
    LIGHTBULB = ("\uE335", "lightbulb")
    RADIATOR_DISABLED = ("\uEAD7", "radiator-disabled")
    LIGHTBULB_ON = ("\uE6E8", "lightbulb-on")
    SNOWFLAKE = ("\uE717", "snowflake")
    OUTDOOR_LAMP = ("\uF054", "outdoor-lamp")
    THERMOMETER = ("\uE50F", "thermometer")
    STRING_LIGHTS = ("\uF2BA", "string-lights")
    WATER = ("\uE58C", "water")
    VANITY_LIGHT = ("\uF1E1", "vanity-light")
    WATER_PERCENT = ("\uE58E", "water-percent")
    WALL_SCONCE = ("\uE91C", "wall-sconce")
    WEATHER_CLOUDY = ("\uE590", "weather-cloudy")
    CHECK = ("\uE12C", "check")
    WEATHER_FOG = ("\uE591", "weather-fog")
    CLOSE = ("\uE156", "close")
    WEATHER_HAIL = ("\uE592", "weather-hail")
    COG = ("\uE493", "cog")
    WEATHER_LIGHTNING = ("\uE593", "weather-lightning")
    DOTS_VERTICAL = ("\uE1D9", "dots-vertical")
    WEATHER_LIGHTNING_RAINY = ("\uE67E", "weather-lightning-rainy")
    HOME = ("\uE2DC", "home")
    WEATHER_NIGHT = ("\uE594", "weather-night")
    HOME_OUTLINE = ("\uE6A1", "home-outline")
    WEATHER_PARTLY_CLOUDY = ("\uE595", "weather-partly-cloudy")
    MINUS = ("\uE374", "minus")
    WEATHER_POURING = ("\uE596", "weather-pouring")
    PLUS = ("\uE415", "plus")
    WEATHER_RAINY = ("\uE597", "weather-rainy")
    BED = ("\uE2E3", "bed")
    WEATHER_SNOWY = ("\uE598", "weather-snowy")
    CAR = ("\uE10B", "car")
    WEATHER_SNOWY_RAINY = ("\uE67F", "weather-snowy-rainy")
    COFFEE = ("\uE176", "coffee")
    WEATHER_SUNNY = ("\uE599", "weather-sunny")
    POOL = ("\uE606", "pool")
    WEATHER_WINDY = ("\uE59D", "weather-windy")
    SHOWER = ("\uE9A0", "shower")
    WEATHER_WINDY_VARIANT = ("\uE59E", "weather-windy-variant")
    SILVERWARE_FORK_KNIFE = ("\uEA70", "silverware-fork-knife")
    WHITE_BALANCE_SUNNY = ("\uE5A8", "white-balance-sunny")
    SOFA = ("\uE4B9", "sofa")
    CALENDAR = ("\uE0ED", "calendar")
    TOILET = ("\uE9AB", "toilet")
    CLOCK_OUTLINE = ("\uE150", "clock-outline")
    ACCOUNT = ("\uE004", "account")
    HISTORY = ("\uE2DA", "history")
    HUMAN_GREETING = ("\uE64A", "human-greeting")
    TIMER_OUTLINE = ("\uE51B", "timer-outline")
    RUN = ("\uE70E", "run")
    BLUETOOTH = ("\uE0AF", "bluetooth")
    ALERT = ("\uE026", "alert")
    WIFI = ("\uE5A9", "wifi")
    CCTV = ("\uE7AE", "cctv")
    BLINDS = ("\uE0AC", "blinds")
    DOOR_CLOSED = ("\uE81B", "door-closed")
    BLINDS_OPEN = ("\uF011", "blinds-open")
    DOOR_CLOSED_LOCK = ("\uF0AF", "door-closed-lock")
    GARAGE_OPEN_VARIANT = ("\uF2D4", "garage-open-variant")
    DOOR_OPEN = ("\uE81C", "door-open")
    GARAGE_VARIANT = ("\uF2D3", "garage-variant")
    KEY_VARIANT = ("\uE30B", "key-variant")
    WINDOW_SHUTTER = ("\uF11C", "window-shutter")
    LOCK = ("\uE33E", "lock")
    WINDOW_SHUTTER_ALERT = ("\uF11D", "window-shutter-alert")
    LOCK_OPEN_VARIANT = ("\uEFC6", "lock-open-variant")
    WINDOW_SHUTTER_OPEN = ("\uF11E", "window-shutter-open")
    SHIELD_CHECK = ("\uE565", "shield-check")
    BELL = ("\uE09A", "bell")
    SHIELD_HOME = ("\uE68A", "shield-home")
    CELLPHONE = ("\uE11C", "cellphone")
    SHIELD_LOCK = ("\uE99D", "shield-lock")
    DISHWASHER = ("\uEAAC", "dishwasher")
    WINDOW_CLOSED_VARIANT = ("\uF1DB", "window-closed-variant")
    ENGINE = ("\uE1FA", "engine")
    MONITOR_SPEAKER = ("\uEF5F", "monitor-speaker")
    FAN = ("\uE210", "fan")
    MUSIC = ("\uE75A", "music")
    FOUNTAIN = ("\uE96B", "fountain")
    PAUSE = ("\uE3E4", "pause")
    FRIDGE_OUTLINE = ("\uE28F", "fridge-outline")
    PLAY = ("\uE40A", "play")
    KETTLE = ("\uE5FA", "kettle")
    REPEAT = ("\uE456", "repeat")
    LAPTOP = ("\uE322", "laptop")
    REPEAT_OFF = ("\uE457", "repeat-off")
    MICROWAVE = ("\uEC99", "microwave")
    REPEAT_ONCE = ("\uE458", "repeat-once")
    RECYCLE_VARIANT = ("\uF39D", "recycle-variant")
    SHUFFLE = ("\uE49D", "shuffle")
    ROBOT_MOWER_OUTLINE = ("\uF1F3", "robot-mower-outline")
    SHUFFLE_DISABLED = ("\uE49E", "shuffle-disabled")
    ROBOT_VACUUM = ("\uE70D", "robot-vacuum")
    SKIP_NEXT = ("\uE4AD", "skip-next")
    STOVE = ("\uE4DE", "stove")
    SKIP_PREVIOUS = ("\uE4AE", "skip-previous")
    TELEVISION = ("\uE502", "television")
    SPEAKER = ("\uE4C3", "speaker")
    TRASH_CAN_OUTLINE = ("\uEA7A", "trash-can-outline")
    STOP = ("\uE4DB", "stop")
    TUMBLE_DRYER = ("\uE917", "tumble-dryer")
    VOLUME_HIGH = ("\uE57E", "volume-high")
    WASHING_MACHINE = ("\uE72A", "washing-machine")
    VOLUME_MEDIUM = ("\uE580", "volume-medium")
    WATER_PUMP = ("\uE58F", "water-pump")
    VOLUME_MUTE = ("\uE75F", "volume-mute")
    ARROW_LEFT_BOX = ("\uE6C1", "arrow-left-box")
    SOLAR_PANEL = ("\uED9B", "solar-panel")
    ARROW_RIGHT_BOX = ("\uE6C2", "arrow-right-box")
    SOLAR_PANEL_LARGE = ("\uED9C", "solar-panel-large")
    RADIATOR_OFF = ("\uEAD8", "radiator-off")
    SOLAR_POWER = ("\uEA72", "solar-power")
    WEATHER_HAZY = ("\uEF30", "weather-hazy")
    SOLAR_POWER_VARIANT_OUTLINE = ("\uFA74", "solar-power-variant-outline")
    WEATHER_NIGHT_PARTLY_CLOUDY = ("\uEF31", "weather-night-partly-cloudy")
    TRANSMISSION_TOWER_IMPORT = ("\uF92D", "transmission-tower-import")
    TRANSMISSION_TOWER = ("\uED3E", "transmission-tower")
    TRANSMISSION_TOWER_EXPORT = ("\uF92C", "transmission-tower-export")

    @classmethod
    def entries(cls) -> List[Tuple[str, str]]:
        """
        Returns a list of tuples containing the unicode value and the name of icon in mdi.
        """
        return [(member.value[0], member.value[1]) for member in cls]


class OpenHaspClient:

    def __init__(self, device: Device):
        """
        :param device: the device this client can communicate with
        """
        self._device = device

        self._image_processor = OpenHaspImageProcessor()

        self._webservice_client = WebserviceClient(
            url=device.config.openhasp_config_manager.device.ip,
            username=self._device.config.http.user,
            password=self._device.config.http.password
        )

        self._mqtt_client = MqttClient(
            host=device.config.mqtt.host,
            port=device.config.mqtt.port,
            mqtt_user=device.config.mqtt.user,
            mqtt_password=device.config.mqtt.password
        )

        self._telnet_client = OpenHaspTelnetClient(
            host=device.config.openhasp_config_manager.device.ip,
            port=device.config.telnet.port,
            baudrate=device.config.debug.baud,
            user=device.config.http.user,
            password=device.config.http.password,
        )

    async def set_text(self, obj: str, text: str):
        """
        Set the text of an object
        :param obj: the object to set the text for
        :param text: the text to set
        """
        await self.set_object_properties(obj, {"text": text})

    async def set_image(
        self,
        obj: str,
        image,
        access_host: str,
        listen_host: str = "0.0.0.0",
        listen_port: int = 0,
        access_port: int = None,
        timeout: int = 10,
        size: Tuple[int or None, int or None] = (None, None),
        fitscreen: bool = False,
    ):
        """
        Sets the image of an object.
        If the image is a URL, it will be fetched first.
        The image will then be converted to RGB565 and
        served by a temporary webserver for the plate to fetch it.

        Note: If you are calling this function concurrently, be aware that under
        the hood a free and unoccupied port is required to feed the image to the
        plate. If you cannot use different ports for each call, make sure to add
        a locking mechanism to avoid multiple threads from creating a webserver
        on the same port.

        :param obj: the object to set the image for
        :param image: the image to set
        :param access_host: the address at which the device this webserver is running on is accessible to the plate
        :param access_port: the port to bind the webserver to, defaults to 0 (random free port)
        :param listen_host: the address to bind the webserver to
        :param timeout: the timeout in seconds to wait for the image to be fetched by the plate
        :param size: the size of the image
        :param fitscreen: if True, the image will be resized to fit the screen
        """
        if access_port is None:
            access_port = listen_port

        import temppathlib
        with temppathlib.NamedTemporaryFile() as out_image:
            self._image_processor.image_to_rgb565(
                in_image=image,
                out_image=out_image.file,
                size=size,
                fitscreen=fitscreen
            )
            await self._serve_image(
                obj=obj,
                image_file=out_image,
                listen_host=listen_host,
                listen_port=listen_port,
                access_host=access_host,
                access_port=access_port,
                timeout=timeout
            )

    async def _serve_image(
        self,
        obj: str,
        image_file,
        listen_host: str,
        access_host: str,
        listen_port: int = 0,
        access_port: int = 0,
        timeout: int = 5,
    ):
        """
        Serves an image using a temporary webserver.

        Note that this webserver intentionally does **not** use HTTPS, since openhasp
        does not support SSL connections. The source file can use an HTTPS source regardless,
        since openhasp-config-manager will fetch the image from the source directly,
        convert it for the plate, and then serve it internally via HTTP.

        :param obj: the object to set the image for
        :param image_file: the image to serve
        :param listen_host: the address to bind the webserver to
        :param listen_port: the port to bind the webserver to, defaults to 0 (random free port)
        :param access_host: the address at which the device this webserver is running on is accessible to the plate
        :param access_port: the port at which the device this webserver is running on is accessible to the plate
        :param timeout: the timeout in seconds after which the webserver will be stopped
        :return: the URL to retrieve the image
        """
        from aiohttp import web

        async def serve_file(request):
            response = web.FileResponse(image_file.path)
            request.app["served"].set_result(True)
            return response

        async def start_server(app, listen_host, listen_port, access_host, access_port, timeout):
            runner = web.AppRunner(app)
            await runner.setup()
            site = web.TCPSite(runner, listen_host, listen_port)
            await site.start()

            port = site._server.sockets[0].getsockname()[1]
            if access_port == 0:
                access_port = port

            listen_url = f"http://{listen_host}:{port}/"
            access_url = f"http://{access_host}:{access_port}/"
            print(f"Serving on {listen_url}, accessible via {access_url}")

            # give the server some time to start
            await asyncio.sleep(1)

            # set the object properties to point to the image url, the plate will download the image immediately
            await self.set_object_properties(
                obj=obj,
                properties={
                    "src": access_url
                }
            )

            try:
                await asyncio.wait_for(app["served"], timeout)
            except asyncio.TimeoutError:
                print(f"Timeout reached after {timeout} seconds. Server stopped.")
            finally:
                await runner.cleanup()

            return listen_url

        app = web.Application()
        app["served"] = asyncio.Future()

        app.router.add_route("GET", "/", serve_file)

        await start_server(
            app=app,
            listen_host=listen_host,
            listen_port=listen_port,
            access_host=access_host,
            access_port=access_port,
            timeout=timeout
        )

    async def set_object_properties(self, obj: str, properties: Dict[str, Any]):
        """
        Set the state of a plate
        :param obj: object to set a state for
        :param properties: properties to set
        """
        for prop, value in properties.items():
            await self.command(
                name=f"{obj}.{prop}",
                params=value,
            )

    async def set_idle_state(self, state: str):
        """
        Forces the given idle state
        :param state: one of "off", "short", "long"
        """
        return await self.command(
            name="idle",
            params=state
        )

    async def set_backlight(self, state: bool, brightness: int):
        """
        Sets the backlight state and brightness
        :param state: True to turn on, False to turn off
        :param brightness: the brightness level in 0..255
        """
        return await self.command(
            name="backlight",
            params={
                "state": state,
                "brightness": brightness,
            }
        )

    async def wakeup(self):
        """
        Wakes up the display
        """
        return await self.set_idle_state(state="off")

    async def set_page(self, index: int):
        """
        Set the current page
        :param index: the index of the page to set
        """
        return await self.command(
            name="page",
            params=f"{index}"
        )

    async def next_page(self):
        """
        Set the next page
        """
        return await self.command(
            name="page",
            params="next"
        )

    async def previous_page(self):
        """
        Set the previous page
        """
        return await self.command(
            name="page",
            params="prev"
        )

    async def clear_current_page(self):
        """
        Clear the current page
        """
        return await self.command(
            name="clearpage",
        )

    async def clear_page(self, page: int):
        """
        Clear the given page
        :param page: the page index to clear
        """
        return await self.command(
            name="clearpage",
            params=page
        )

    async def clear_all_pages(self):
        """
        Clear all pages
        """
        return await self.command(
            name="clearpage",
            params="all"
        )

    async def set_hidden(self, obj: str, hidden: bool):
        """
        Set the "hidden" property of an object
        :param obj: the object to set the property for
        :param hidden: True to hide the object, False to show it
        """
        return await self.set_object_properties(
            obj=obj,
            properties={
                "hidden": "1" if hidden else "0",
            }
        )

    async def listen_state(self, obj: str, callback: Callable):
        """
        Listen to OpenHASP state events
        :param obj: object to listen to
        :param callback: callback to call when a matching event is received
        """

        async def _callback(event_topic: Topic, event_payload: bytes):
            await callback(str(event_topic), event_payload)

        await self.listen_event(
            path=f"state/{obj}",
            callback=_callback,
        )

    async def listen_event(self, path: str, callback: Callable):
        """
        Listen to OpenHASP events related to this device
        :param path: MQTT subpath to listen to
        :param callback: callback to call when a matching event is received
        """
        topic = f"hasp/{self._device.config.mqtt.name}/{path}"
        await self._mqtt_client.subscribe(topic, callback)

    async def cancel_callback(self, callback: Callable = None):
        """
        Cancel a callback which was previously registered, f.ex. via listen_event (or listen_state)
        :param callback: the specific callback to cancel
        """
        await self._mqtt_client.cancel_callback(callback=callback)

    async def command(self, name: str, params: any):
        """
        Execute a command on a device
        :param name: the name of the command
        :param params: parameters for the command
        """
        topic = f"hasp/{self._device.config.mqtt.name}/command/{name}"
        await self._mqtt_client.publish(topic=topic, payload=params)

    def get_files(self) -> List[str]:
        """
        Retrieve a list of all file on the device
        :return: a list of all files on the device
        """
        return self._webservice_client.get_files()

    def get_file_content(self, file_name: str) -> bytes or None:
        return self._webservice_client.get_file_content(file_name)

    def delete_file(self, file_name: str):
        """
        Delete a file on the device
        :param file_name: the name of the file
        """
        self._webservice_client.delete_file(file_name)

    def reboot(self):
        """
        Request a reboot
        """
        self._webservice_client.reboot()

    def set_hasp_config(self, config: HaspConfig):
        """
        Set the "HASP" configuration
        :param config: the configuration to set
        """
        self._webservice_client.set_hasp_config(config)

    def get_http_config(self) -> HttpConfig:
        """
        Get the HTTP configuration
        :return: the HTTP configuration
        """
        return self._webservice_client.get_http_config()

    def set_http_config(self, config: HttpConfig):
        """
        Set the HTTP configuration
        :param config: the configuration to set
        """
        self._webservice_client.set_http_config(config)

    def get_mqtt_config(self) -> MqttConfig:
        return self._webservice_client.get_mqtt_config()

    def set_mqtt_config(self, config: MqttConfig):
        """
        Set the MQTT configuration
        :param config: the configuration to set
        :return:
        """
        self._webservice_client.set_mqtt_config(config)

    def get_gui_config(self) -> GuiConfig:
        """
        Get the GUI configuration
        :return: the GUI configuration
        """
        return self._webservice_client.get_gui_config()

    def set_gui_config(self, config: GuiConfig):
        """
        Set the GUI configuration
        :param config: the configuration to set
        :return:
        """
        self._webservice_client.set_gui_config(config)

    def upload_files(self, files: Dict[str, bytes]):
        """
        Upload a collection of files
        :param files: "target file name"->"file content" mapping
        """
        self._webservice_client.upload_files(files)

    def upload_file(self, name: str, content: bytes):
        """
        Upload a single file
        :param name: the target name of the file on the device
        :param content: the file content
        """
        info(f"Uploading '{name}'...")
        self._webservice_client.upload_file(name, content)

    def take_screenshot(self) -> bytes:
        """
        Requests a screenshot from the device.
        :return:
        """
        return self._webservice_client.take_screenshot()

    async def shell(self):
        await self._telnet_client.shell()

    async def logs(self):
        await self._telnet_client.logs()



import abc
from typing import Dict

from openhasp_config_manager.openhasp_client.model.configuration.config import Config


class JsonlObjectProcessor(metaclass=abc.ABCMeta):
    """
    Used to process .jsonl files to add support for additional features.
    """

    @abc.abstractmethod
    def process(self, input: Dict, config: Config, template_vars: Dict[str, any]) -> Dict:
        raise NotImplementedError



from dataclasses import dataclass

from openhasp_config_manager.openhasp_client.model.configuration.screen_config import ScreenConfig


@dataclass
class DeviceConfig:
    ip: str
    screen: ScreenConfig



import pathlib
from pathlib import Path

from PyQt6.QtWidgets import QTreeWidget, QAbstractItemView, QTreeWidgetItem, QSizePolicy

from openhasp_config_manager.ui.qt.util import clear_layout


class FileBrowserWidget(QTreeWidget):
    def __init__(self, cfg_root: Path):
        super().__init__()
        self.cfg_root = cfg_root
        self.create_layout()
        self.create_entries()

    def create_layout(self):
        self.setHeaderLabel("File Browser")
        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setAnimated(False)
        self.setAllColumnsShowFocus(True)

        self.itemClicked.connect(self.on_item_clicked)

        size_policy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        self.setSizePolicy(size_policy)

    def clear_entries(self):
        clear_layout(self.layout)

    def create_entries(self):
        self._load_file_structure(self.cfg_root, self)

    def _load_file_structure(self, startpath: Path, tree: QTreeWidget):
        """
        Load Project structure tree
        :param startpath:
        :param tree:
        :return:
        """
        import os
        from PyQt6.QtWidgets import QTreeWidgetItem
        from PyQt6.QtGui import QIcon
        startpath.is_dir()
        for element in sorted(pathlib.Path(startpath).iterdir(), key=lambda x: (not x.is_dir(), x.name)):
            path_info = Path(startpath, element)
            parent_itm = QTreeWidgetItem(tree, [os.path.basename(element)])
            if os.path.isdir(path_info):
                self._load_file_structure(path_info, parent_itm)
                parent_itm.setIcon(0, QIcon('assets/folder.ico'))
            else:
                parent_itm.setIcon(0, QIcon('assets/file.ico'))

    def on_item_clicked(self, it: QTreeWidgetItem, col: int):
        print(it, col, it.text(col))



from pathlib import Path

from openhasp_config_manager.cli.common import _create_config_manager, _analyze_and_filter, _cmd
from openhasp_config_manager.ui.util import success, error


async def c_cmd(config_dir: Path, device: str, command: str, payload: str):
    try:
        config_manager = _create_config_manager(config_dir, Path("./nonexistent"))
        filtered_devices, ignored_devices = _analyze_and_filter(config_manager=config_manager, device_filter=device)

        if len(filtered_devices) <= 0:
            raise Exception(f"No device matches the filter: {device}")

        for device in filtered_devices:
            await _cmd(device, command, payload)
        success("Done!")
    except Exception as ex:
        error(str(ex))



import json
from typing import List, Dict, Any

import orjson

from openhasp_config_manager.openhasp_client.model.component import Component, JsonlComponent, CmdComponent, \
    TextComponent, RawComponent
from openhasp_config_manager.openhasp_client.model.configuration.config import Config
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.processing.jsonl import JsonlObjectProcessor
from openhasp_config_manager.processing.preprocessor.jsonl_preprocessor import JsonlPreProcessor
from openhasp_config_manager.processing.template_rendering import render_dict_recursive, _render_template
from openhasp_config_manager.processing.variables import VariableManager
from openhasp_config_manager.util import merge_dict_recursive


class DeviceProcessor:
    """
    A device specific processor, used to transform any templates
    present within the configuration files.
    """

    def __init__(self, device: Device, jsonl_object_processors: List[JsonlObjectProcessor],
                 variable_manager: VariableManager):
        self._device = device

        self._jsonl_components: List[JsonlComponent] = []
        self._others: List[Component] = []

        self._jsonl_preprocessor = JsonlPreProcessor()
        self._jsonl_object_processors = jsonl_object_processors
        self._variable_manager = variable_manager

    def add_component(self, component: Component):
        if isinstance(component, JsonlComponent):
            self._add_jsonl(component)
        else:
            self._add_other(component)

    def _add_other(self, component: Component):
        self._others.append(component)

    def _add_jsonl(self, component: JsonlComponent):
        self._jsonl_components.append(component)

    def normalize(self, device: Device, component: Component) -> str | bytes:
        if isinstance(component, JsonlComponent):
            template_vars: Dict[str, any] = self._compute_jsonl_template_variables(device, component)
            return self._normalize_jsonl(self._device.config, component, template_vars)
        elif isinstance(component, CmdComponent):
            template_vars: Dict[str, any] = {}
            return self._normalize_cmd(self._device.config, component, template_vars)
        elif isinstance(component, TextComponent):
            # no changes necessary
            return component.content
        elif isinstance(component, RawComponent):
            # no changes necessary
            return component.content
        elif isinstance(component, Component):
            raise NotImplementedError(f"Unknown component type: {component.type}")
        else:
            raise AssertionError(f"Received unexpected input: {component}")

    def _normalize_jsonl(self, config: Config, component: JsonlComponent, template_vars: Dict[str, any]) -> str:
        normalized_objects: List[str] = []

        objects = self._jsonl_preprocessor.split_jsonl_objects(component.content)
        for ob in objects:
            preprocessed = self._jsonl_preprocessor.cleanup_object_for_json_parsing(ob)
            p = self._normalize_jsonl_object(config, component, preprocessed, template_vars)
            normalized_objects.append(p)

        return "\n".join(normalized_objects)

    def _normalize_jsonl_object(
        self, config: Config, component: JsonlComponent, ob: str, template_vars: Dict[str, any]
    ) -> str:
        parsed = orjson.loads(ob)

        normalized_object = {}
        for key, value in parsed.items():
            if isinstance(value, str):
                rendered_value = _render_template(value, template_vars)
                normalized_object[key] = rendered_value
            else:
                normalized_object[key] = value

        processed = normalized_object
        for processor in self._jsonl_object_processors:
            processed = processor.process(processed, config, template_vars)

        return json.dumps(processed, indent=None)

    def _normalize_cmd(self, _device_config, component: CmdComponent, template_vars: Dict[str, any]) -> str:
        return _render_template(component.content, template_vars)

    def _compute_jsonl_template_variables(self, device: Device, component: Component) -> Dict[str, Any]:
        """
        Computes a map of "variable" -> "evaluated value in the given path context" for the given component.

        :param component: the component to use as a context for evaluating template variables
        :return: map of "variable" -> "evaluated value in the given path context"
        """
        result = {}

        # object specific variables for all components that the processor currently knows about
        sorted_components = list(sorted(self._jsonl_components, key=lambda x: x == component))
        assert sorted_components[-1] == component
        for c in sorted_components:
            jsonl_objects = self._jsonl_preprocessor.split_jsonl_objects(c.content)

            component_template_vars = self._variable_manager.get_vars(c.path)
            component_template_vars["device"] = self._device.config.openhasp_config_manager.device
            if "/common/" in str(c.path):
                # for common components, also include device specific, top-level variables
                device_vars = self._variable_manager.get_vars(device.path)
                component_template_vars = merge_dict_recursive(component_template_vars, device_vars)

            c_result = self._compute_object_map(jsonl_objects)
            merged_template_variables = merge_dict_recursive(result, component_template_vars)
            merged_template_variables = merge_dict_recursive(merged_template_variables, component_template_vars)

            rendered_template_vars = render_dict_recursive(
                input=c_result,
                template_vars=merge_dict_recursive(merged_template_variables, c_result)
            )
            rendered_template_vars = merge_dict_recursive(rendered_template_vars, merged_template_variables)

            result = merge_dict_recursive(result, rendered_template_vars)

        if result is None:
            raise AssertionError("Unexpected None value")
        return result

    def _compute_object_map(self, jsonl_objects: List[str]) -> Dict[str, dict]:
        """
        :param jsonl_objects:
        :return: a map
        """
        result = {}
        for o in jsonl_objects:
            o = self._jsonl_preprocessor.cleanup_object_for_json_parsing(o)
            parsed_object = orjson.loads(o)
            object_key = self._compute_object_key(parsed_object)
            result[object_key] = parsed_object
        return result

    @staticmethod
    def _compute_object_key(jsonl_object: Dict) -> str:
        return f"p{jsonl_object.get('page', '0')}b{jsonl_object.get('id', '0')}"



import sys
from pathlib import Path

from PyQt6.QtWidgets import QApplication

from openhasp_config_manager.cli.common import _create_config_manager
from openhasp_config_manager.ui.qt.main import MainWindow


async def c_gui(config_dir: Path, output_dir: Path):
    config_manager = _create_config_manager(config_dir, output_dir)
    app = QApplication(sys.argv)

    window = MainWindow(config_manager)
    window.show()

    app.exec()



from typing import Dict

import orjson as orjson
from py_range_parse import Range

from openhasp_config_manager.validation import Validator


class JsonlObjectValidator(Validator):
    """
    Validates JSONL object definitions.
    """

    def __init__(self):
        self._seen_ids = {}

    def validate(self, data: str):
        for line in data.splitlines():
            input = orjson.loads(line)
            self._validate_object(input)

    def _validate_object(self, input: Dict):
        input_page = input.get("page", None)
        input_id = input.get("id", None)
        self.__remember_page_id_combo(input_page, input_id, input)

        if input_id is not None:
            valid_range = Range(0, 254)
            if not isinstance(input_id, int) or input_id not in valid_range:
                raise AssertionError(f"Object has invalid id '{input_id}', must be in range: {valid_range}")

        input_align = input.get("align", None)
        if input_align is not None:
            valid_align_int_values = [0, 1, 2]
            valid_align_str_values = ["left", "center", "right"]
            valid_align_values = valid_align_str_values + valid_align_int_values

            if input_align not in valid_align_values:
                if isinstance(input_align, str):
                    raise AssertionError(
                        f"Invalid 'align' string value: '{input_align}', must be one of: {valid_align_values}")
                if isinstance(input_align, int):
                    raise AssertionError(
                        f"Invalid 'align' integer value: '{input_align}', must be one of: {valid_align_values}")

    def __remember_page_id_combo(self, input_page: int, input_id: int, data: Dict):
        if input_page is None or input_id is None:
            raise AssertionError(f"page or id is None: {input_page}, {input_id}")

        key = f"p{input_page}b{input_id}"
        if key in self._seen_ids.keys():
            raise AssertionError(f"Duplicate id detected: {key}, already seen in object: {self._seen_ids[key]}")
        else:
            self._seen_ids[key] = data



from dataclasses import dataclass
from typing import List


@dataclass
class GuiConfig:
    idle1: int
    idle2: int
    bckl: int
    bcklinv: int
    rotate: int
    cursor: int
    invert: int
    calibration: List[int]



from openhasp_config_manager.openhasp_client.model.component import Component
from openhasp_config_manager.openhasp_client.model.configuration.config import Config
from openhasp_config_manager.validation.cmd import CmdFileValidator
from openhasp_config_manager.validation.jsonl import JsonlObjectValidator


class DeviceValidator:
    """
    A device specific validator, used to validate the result of the DeviceProcessor.
    """

    def __init__(self, config: Config, jsonl_object_validator: JsonlObjectValidator,
                 cmd_file_validator: CmdFileValidator):
        self._config = config
        self._jsonl_object_validator = jsonl_object_validator
        self._cmd_file_validator = cmd_file_validator

    def validate(self, component: Component, data: str):
        if component.type == "jsonl":
            self._jsonl_object_validator.validate(data)
        if component.type == "cmd":
            self._cmd_file_validator.validate(data)
        if len(component.name) > 30:
            raise AssertionError(
                f"Output file name length must not exceed 30 characters, but was {len(component.name)}: {component.name}")



from typing import Dict, List

import orjson
import requests

from openhasp_config_manager.openhasp_client.model.configuration.debug_config import DebugConfig
from openhasp_config_manager.openhasp_client.model.configuration.gui_config import GuiConfig
from openhasp_config_manager.openhasp_client.model.configuration.hasp_config import HaspConfig
from openhasp_config_manager.openhasp_client.model.configuration.http_config import HttpConfig
from openhasp_config_manager.openhasp_client.model.configuration.mqtt_config import MqttConfig, MqttTopicConfig
from openhasp_config_manager.openhasp_client.model.configuration.telnet_config import TelnetConfig

GET = "GET"
POST = "POST"
DELETE = "DELETE"


class WebserviceClient:

    def __init__(self, url: str, username: str, password: str):
        self._username = username
        self._password = password
        self._base_url = self._compute_base_url(url)

    def reboot(self):
        """
        Request a reboot
        """
        self._do_request(
            method=GET,
            url=self._base_url + "reboot",
        )

    def set_hasp_config(self, config: HaspConfig):
        """
        Set the "HASP" configuration
        :param config: the configuration to set
        """
        data = {
            "startpage": config.startpage,
            "startdim": config.startdim,
            "theme": config.theme,
            "color1": config.color1,
            "color2": config.color2,
            "font": config.font,
            "pages": config.pages,
            "save": "hasp"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def get_http_config(self) -> HttpConfig:
        data = self._do_request(
            method=GET,
            url=self._base_url + "api/config/http/",
        )

        return HttpConfig(
            port=data["port"],
            user=data["user"],
            password=data["pass"],
        )

    def set_http_config(self, config: HttpConfig):
        """
        Set the HTTP configuration
        :param config: the configuration to set
        """
        data = {
            "user": config.user,
            "pass": config.password,
            "save": "http"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def get_mqtt_config(self) -> MqttConfig:
        data = self._do_request(
            method=GET,
            url=self._base_url + "api/config/mqtt/",
        )

        return MqttConfig(
            name=data["name"],
            host=data["host"],
            port=data["port"],
            user=data["user"],
            password=data["pass"],
            topic=MqttTopicConfig(
                node=data["topic"]["node"],
                group=data["topic"]["group"],
                broadcast=data["topic"]["broadcast"],
                hass=data["topic"]["hass"],
            )
        )

    def set_mqtt_config(self, config: MqttConfig):
        """
        Set the MQTT configuration
        :param config: the configuration to set
        """
        data = {
            "name": config.name,
            "topic": {
                "node": config.topic.node,
                "group": config.topic.group,
                "broadcast": config.topic.broadcast,
                "hass": config.topic.hass,
            },
            "host": config.host,
            "port": config.port,
            "user": config.user,
            "pass": config.password,
            "save": "mqtt"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def get_gui_config(self) -> GuiConfig:
        data = self._do_request(
            method=GET,
            url=self._base_url + "api/config/gui/",
        )

        return GuiConfig(
            idle1=data["idle1"],
            idle2=data["idle2"],
            rotate=data["rotate"],
            cursor=data["cursor"],
            bckl=data["bckl"],
            bcklinv=data["bcklinv"],
            invert=data["invert"],
            calibration=data["calibration"]
        )

    def set_gui_config(self, config: GuiConfig):
        """
        Set the GUI configuration
        :param config: the configuration to set
        """
        data = {
            "idle1": config.idle1,
            "idle2": config.idle2,
            "rotate": config.rotate,
            "cursor": config.cursor,
            "bckl": config.bckl,
            "save": "gui"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def set_telnet_config(self, config: TelnetConfig):
        """
        Set the Debug configuration
        :param config: the configuration to set
        """
        data = {
            "enable": config.enable,
            "port": config.port,
            "save": "telnet"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def set_debug_config(self, config: DebugConfig):
        """
        Set the Debug configuration
        :param config: the configuration to set
        """
        data = {
            "ansi": config.ansi,
            "baud": config.baud,
            "tele": config.tele,
            "host": config.host,
            "port": config.port,
            "proto": config.proto,
            "log": config.log,
            "save": "debug"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        self._do_request(
            method=POST,
            url=self._base_url + "config",
            data=data,
        )

    def upload_files(self, files: Dict[str, bytes]):
        """
        Upload a collection of files
        :param files: "target file name"->"file content" mapping
        """
        for name, content in files.items():
            self.upload_file(name, content)

    def upload_file(self, name: str, content: bytes):
        """
        Upload a single file
        :param name: the target name of the file on the device
        :param content: the file content
        """
        self._do_request(
            method=POST,
            url=self._base_url + "edit",
            files={
                f"{name}": content
            },
        )

    def get_files(self) -> List[str]:
        """
        Retrieve a list of all file on the device
        :return: a list of all files on the device
        """
        response = self._do_request(
            method=GET,
            url=self._base_url + "list?dir=/",
        )
        response_data = orjson.loads(response.decode('utf-8'))

        files = list(filter(lambda x: x["type"] == "file", response_data))
        file_names = list(map(lambda x: x["name"], files))
        return file_names

    def get_file_content(self, file_name: str) -> bytes or None:
        try:
            response = self._do_request(
                method=GET,
                url=self._base_url + file_name,
            )
            response_data = response
            return response_data
        except Exception as ex:
            return None

    def delete_file(self, file_name: str):
        """
        Delete a file on the device
        :param file_name: the name of the file
        """
        self._do_request(
            method=DELETE,
            url=self._base_url + "edit",
            data={
                "path": "/" + file_name
            },
        )

    def take_screenshot(self) -> bytes:
        """
        Requests a screenshot from the device.
        :return:
        """
        return self._do_request(
            method=GET,
            url=self._base_url + "screenshot",
            params={
                "q": "0"
            },
            stream=True,
        )

    def _do_request(self, method: str = GET, url: str = "/", params: dict = None,
                    json: any = None, files: any = None, data: any = None,
                    headers: dict = None,
                    stream: bool = None) -> list or dict or None:
        """
        Executes a http request based on the given parameters

        :param method: the method to use (GET, POST)
        :param url: the url to use
        :param params: query parameters that will be appended to the url
        :param json: request body
        :param headers: custom headers
        :return: the response parsed as a json
        """
        _headers = {
        }
        if headers is not None:
            _headers.update(headers)

        response = requests.request(
            method, url, headers=_headers,
            params=params,
            json=json, files=files,
            data=data,
            auth=(self._username, self._password),
            timeout=5,
            stream=stream,
        )

        response.raise_for_status()
        if len(response.content) > 0:
            if "application/json" in response.headers.get("Content-Type", ""):
                return response.json()
            else:
                return response.content

    @staticmethod
    def _compute_base_url(url: str) -> str:
        if not url.startswith("http://"):
            url = "http://" + url
        if not url.endswith("/"):
            url += "/"
        return url



from openhasp_config_manager.cli import cli


def main():
    import os
    import sys

    parent_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), "..", ".."))
    sys.path.append(parent_dir)

    cli()


if __name__ == '__main__':
    main()



from dataclasses import dataclass


@dataclass
class MqttConfig:
    name: str
    host: str
    port: int
    user: str
    password: str
    topic: "MqttTopicConfig"


@dataclass
class MqttTopicConfig:
    node: str
    group: str
    broadcast: str
    hass: str



import re
import shutil
from pathlib import Path
from typing import List, Set

import orjson

from openhasp_config_manager.const import COMMON_FOLDER_NAME, DEVICES_FOLDER_NAME, SYSTEM_SCRIPTS
from openhasp_config_manager.openhasp_client.model.component import Component, TextComponent, RawComponent, \
    ImageComponent, JsonlComponent, CmdComponent, FontComponent
from openhasp_config_manager.openhasp_client.model.configuration.config import Config
from openhasp_config_manager.openhasp_client.model.configuration.debug_config import DebugConfig
from openhasp_config_manager.openhasp_client.model.configuration.device_config import DeviceConfig
from openhasp_config_manager.openhasp_client.model.configuration.gui_config import GuiConfig
from openhasp_config_manager.openhasp_client.model.configuration.hasp_config import HaspConfig
from openhasp_config_manager.openhasp_client.model.configuration.http_config import HttpConfig
from openhasp_config_manager.openhasp_client.model.configuration.mqtt_config import MqttConfig, MqttTopicConfig
from openhasp_config_manager.openhasp_client.model.configuration.screen_config import ScreenConfig
from openhasp_config_manager.openhasp_client.model.configuration.telnet_config import TelnetConfig
from openhasp_config_manager.openhasp_client.model.configuration.wifi_config import WifiConfig
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.openhasp_client.model.openhasp_config_manager_config import OpenhaspConfigManagerConfig
from openhasp_config_manager.processing.device_processor import DeviceProcessor
from openhasp_config_manager.processing.jsonl.jsonl import ObjectDimensionsProcessor, ObjectThemeProcessor
from openhasp_config_manager.processing.variables import VariableManager
from openhasp_config_manager.ui.util import warn
from openhasp_config_manager.validation.cmd import CmdFileValidator
from openhasp_config_manager.validation.device_validator import DeviceValidator
from openhasp_config_manager.validation.jsonl import JsonlObjectValidator

CONFIG_FILE_NAME = "config.json"


def parse_cmd_commands(content) -> List[str]:
    """
    Parses the commands from the given cmd component content.
    :param content: the content of the cmd component
    :return: list of commands
    """
    result = []
    for line in content.splitlines():
        line = line.strip()
        if len(line) <= 0:
            continue
        if line.startswith("//"):
            continue
        result.append(line)
    return result


class ConfigManager:
    """
    This class is used to manage multiple device configurations that may be present
    within a given config directory.
    """

    def __init__(self, cfg_root: Path, output_root: Path, variable_manager: VariableManager):
        self.cfg_root = cfg_root
        self._output_root = output_root

        self._variable_manager = variable_manager

    def analyze(self) -> List[Device]:
        """
        Analyze the configuration file tree and the contents of relevant files.
        The result will be a list of Device object instances, representing the structure
        and configuration found.

        :return: list of devices, which can be used in the ConfigManager.process() method
        """
        self._variable_manager.read()
        return self._analyze(self.cfg_root, self._output_root)

    def _analyze(self, cfg_dir_root: Path, output_dir_root: Path) -> List[Device]:
        result: List[Device] = []

        common_components_path = Path(cfg_dir_root, COMMON_FOLDER_NAME)
        common_components = self._read_components(common_components_path, prefix="common")

        devices_path = Path(cfg_dir_root, DEVICES_FOLDER_NAME)
        if not devices_path.exists():
            raise RuntimeError(
                f"No '{DEVICES_FOLDER_NAME}' sub-folder found in '{cfg_dir_root}'. Please create it and move your "
                f"device configuration files there.")
        for device_path in devices_path.iterdir():
            if not device_path.is_dir():
                continue

            device_output_dir = Path(output_dir_root, device_path.name)
            try:
                config = self._read_config(device_path)
            except Exception as ex:
                raise Exception(f"Error reading config '{device_path}': {ex}")

            device_components = self._analyze_device(config, device_path)

            combined_components = common_components + device_components

            device = Device(
                path=device_path,
                name=device_path.name,
                jsonl=[c for c in combined_components if isinstance(c, JsonlComponent)],
                cmd=[c for c in combined_components if isinstance(c, CmdComponent)],
                images=[c for c in combined_components if isinstance(c, ImageComponent)],
                fonts=[c for c in combined_components if isinstance(c, FontComponent)],
                config=config,
                output_dir=device_output_dir,
            )

            result.append(device)

        return result

    def _analyze_device(self, config: Config, device_cfg_dir_root: Path) -> List[Component]:
        result = self._read_components(device_cfg_dir_root)
        return result

    @staticmethod
    def _compute_component_path_of_pages_config_value(device_cfg_dir_root: Path, config: Config) -> Path | None:
        sub_path = config.hasp.pages[1:]
        sub_path = sub_path.removeprefix("L:/")
        sub_path = sub_path.removeprefix("/")
        if len(sub_path) <= 0:
            return None
        path = Path(device_cfg_dir_root, sub_path)
        relative_path = path.relative_to(device_cfg_dir_root)
        if path.is_file():
            return relative_path
        else:
            warn(f"Could not find file '{relative_path}' referenced by hasp.pages config property, searched in: {path}")
            return None

    def _read_components(self, path: Path, prefix: str = "") -> List[Component]:
        """
        Recursively reads all components from the given path.
        :param path: the root path to read components from
        :param prefix: a prefix to add to the component name
        :return: list of components
        """
        result: List[Component] = []

        result.extend(self._read_jsonl_components(path, prefix))
        result.extend(self._read_cmd_components(path, prefix))
        result.extend(self._read_image_components(path, prefix))

        return result

    def _read_jsonl_components(self, path, prefix) -> List[Component]:
        result = []
        suffix = ".jsonl"
        for file in path.rglob(f"*{suffix}"):
            component = self._create_text_component_from_path(
                device_cfg_dir_root=path,
                path=Path(path, file.relative_to(path)).relative_to(path),
                prefix=prefix
            )
            if component is not None:
                jsonl_component = JsonlComponent(
                    name=component.name,
                    type=component.type,
                    path=component.path,
                    content=component.content,
                )
                result.append(jsonl_component)
        return result

    def _read_cmd_components(self, path, prefix) -> List[Component]:
        result = []
        suffix = ".cmd"
        for file in path.rglob(f"*{suffix}"):
            component = self._create_text_component_from_path(
                device_cfg_dir_root=path,
                path=Path(path, file.relative_to(path)).relative_to(path),
                prefix=prefix
            )
            if component is not None:
                cmd_component = CmdComponent(
                    name=component.name,
                    type=component.type,
                    path=component.path,
                    content=component.content,
                    commands=parse_cmd_commands(component.content)
                )
                result.append(cmd_component)
        return result

    def _read_image_components(self, path, prefix) -> List[Component]:
        result = []
        image_suffixes = [".png", ".bin"]
        for suffix in image_suffixes:
            for file in path.rglob(f"*{suffix}"):
                component = self._create_raw_component_from_path(
                    device_cfg_dir_root=path,
                    path=Path(path, file.relative_to(path)).relative_to(path),
                    prefix=prefix
                )
                if component is not None:
                    image_component = ImageComponent(
                        name=component.name,
                        type=component.type,
                        path=component.path,
                        content=component.content,
                    )
                    result.append(image_component)
        return result

    @staticmethod
    def _create_text_component_from_path(device_cfg_dir_root: Path, path: Path, prefix: str = "") -> Component or None:
        file = Path(device_cfg_dir_root, path)

        if not file.is_file():
            warn(f"Not a file, skipping: {file}")
            return None

        content = file.read_text()

        name_parts = []
        if len(prefix) > 0:
            name_parts.append(prefix)
        name_parts = name_parts + list(file.relative_to(device_cfg_dir_root).parts)

        name = "_".join(name_parts)
        suffix = file.suffix
        component = TextComponent(
            name=name,
            type=suffix[1:],
            path=file,
            content=content,
        )
        return component

    @staticmethod
    def _create_raw_component_from_path(device_cfg_dir_root: Path, path: Path, prefix: str) -> Component or None:
        file = Path(device_cfg_dir_root, path)

        if not file.is_file():
            warn(f"Not a file, skipping: {file}")
            return None

        content = file.read_bytes()

        name_parts = []
        if len(prefix) > 0:
            name_parts.append(prefix)
        name_parts = name_parts + list(file.relative_to(device_cfg_dir_root).parts)

        name = "_".join(name_parts)
        suffix = file.suffix
        component = RawComponent(
            name=name,
            type=suffix[1:],
            path=file,
            content=content,
        )
        return component

    def _read_config(self, device_path: Path) -> Config | None:
        config_file = Path(device_path, CONFIG_FILE_NAME)
        if config_file.exists() and config_file.is_file():
            content = config_file.read_text()
            loaded = orjson.loads(content)

            gui_config = self._parse_gui_config(loaded["gui"])
            is_screen_rotated = gui_config.rotate % 2 == 1

            config = Config(
                openhasp_config_manager=self._parse_openhasp_config_manager_config(
                    data=loaded["openhasp_config_manager"],
                    swap_width_and_height=is_screen_rotated
                ),
                wifi=self._parse_wifi_config(loaded["wifi"]),
                mqtt=self._parse_mqtt_config(loaded["mqtt"]),
                http=self._parse_http_config(loaded["http"]),
                gui=self._parse_gui_config(loaded["gui"]),
                hasp=self._parse_hasp_config(loaded["hasp"]),
                debug=self._parse_debug_config(loaded["debug"]),
                telnet=self._parse_telnet_config(loaded["telnet"])
            )

            return config

    @staticmethod
    def _parse_openhasp_config_manager_config(data: dict, swap_width_and_height: bool) -> OpenhaspConfigManagerConfig:
        screen_width_key = "width"
        screen_height_key = "height"

        if swap_width_and_height:
            # swap width and height if screen is rotated
            t = screen_width_key
            screen_width_key = screen_height_key
            screen_height_key = t

        return OpenhaspConfigManagerConfig(
            device=DeviceConfig(
                ip=data["device"]["ip"],
                screen=ScreenConfig(
                    width=data["device"]["screen"][screen_width_key],
                    height=data["device"]["screen"][screen_height_key]
                )
            )
        )

    @staticmethod
    def _parse_wifi_config(data: dict) -> WifiConfig:
        return WifiConfig(
            ssid=data["ssid"],
            password=data["pass"],
        )

    @staticmethod
    def _parse_mqtt_config(data: dict) -> MqttConfig:
        return MqttConfig(
            name=data["name"],
            topic=MqttTopicConfig(
                node=data["topic"]["node"],
                group=data["topic"]["group"],
                broadcast=data["topic"]["broadcast"],
                hass=data["topic"]["hass"],
            ),
            host=data["host"],
            port=data["port"],
            user=data["user"],
            password=data["pass"],
        )

    @staticmethod
    def _parse_http_config(data: dict) -> HttpConfig:
        return HttpConfig(
            port=data["port"],
            user=data["user"],
            password=data["pass"],
        )

    @staticmethod
    def _parse_gui_config(data: dict) -> GuiConfig:
        return GuiConfig(
            idle1=data["idle1"],
            idle2=data["idle2"],
            bckl=data["bckl"],
            bcklinv=data["bcklinv"],
            rotate=data["rotate"],
            cursor=data["cursor"],
            invert=data["invert"],
            calibration=data["calibration"],
        )

    @staticmethod
    def _parse_hasp_config(data: dict) -> HaspConfig:
        return HaspConfig(
            startpage=data["startpage"],
            startdim=data["startdim"],
            theme=data["theme"],
            color1=data["color1"],
            color2=data["color2"],
            font=data["font"],
            pages=data["pages"],
        )

    @staticmethod
    def _parse_debug_config(data: dict) -> DebugConfig:
        return DebugConfig(
            ansi=data["ansi"],
            baud=data["baud"],
            tele=data["tele"],
            host=data["host"],
            port=data["port"],
            proto=data["proto"],
            log=data["log"],
        )

    @staticmethod
    def _parse_telnet_config(data: dict) -> TelnetConfig:
        return TelnetConfig(
            enable=data["enable"],
            port=data["port"],
        )

    def process(self, device: Device):
        """
        Process the configuration and generate the corresponding output
        """
        self._generate_output(device)

    def _generate_output(self, device: Device):
        self._clear_output(device)

        device_processor = self.create_device_processor(device)
        device_validator = self.create_device_validator(device)

        # only include files which are referenced in a cmd file
        relevant_components = self.find_relevant_components(device)

        # let the processor manage each component
        for component in relevant_components:
            try:
                output_content = device_processor.normalize(device, component)
            except Exception as ex:
                raise Exception(f"Error normalizing {component.path}: {ex}")

            try:
                device_validator.validate(component, output_content)
            except Exception as ex:
                raise Exception(f"Validation for {component.path} failed: {ex}")

            self._write_output(device, component, output_content)

    def create_device_processor(self, device: Device):
        """
        Creates a DeviceProcessor for the given device.
        :param device: the device to create the processor for
        :return: a DeviceProcessor instance
        """
        # prepare DeviceProcessor
        jsonl_processors = [
            ObjectDimensionsProcessor(),
            ObjectThemeProcessor()
        ]
        device_processor = DeviceProcessor(device, jsonl_processors, self._variable_manager)

        # feed device specific data to the processor
        # Note: this also includes common components
        components: List[Component] = device.jsonl + device.cmd + device.images + device.fonts
        for component in components:
            device_processor.add_component(component)

        return device_processor

    def create_device_validator(self, device: Device):
        """
        Creates a DeviceValidator for the given device.
        :param device: the device to create the validator for
        :return: a DeviceValidator instance
        """
        # prepare DeviceValidator
        jsonl_validator = JsonlObjectValidator()
        cmd_file_validator = CmdFileValidator()
        device_validator = DeviceValidator(device.config, jsonl_validator, cmd_file_validator)
        return device_validator

    def find_relevant_components(self, device: Device) -> List[Component]:
        """
        Find all components that are relevant for the given device.
        This includes all components that are referenced in the CMD files,
        as well as all components that are referenced in the JSONL files.

        TODO: allow the user to explicitly include specific components

        :param device: the device to find components for
        :return: a list of relevant components
        """
        result: List[Component] = []
        cmd_components = device.cmd
        jsonl_components = device.jsonl
        image_components = device.images

        referenced_cmd_components = self._find_referenced_cmd_components(cmd_components)

        # find jsonl files referenced in CMD files
        referenced_jsonl_components = self._find_referenced_jsonl_components(cmd_components, jsonl_components)

        # find image files referenced in CMD files and JSONL files
        referenced_image_components = self._find_referenced_image_components(
            cmd_components, jsonl_components, image_components
        )

        # compute component for jsonl file referenced in config.hasp.pages
        pages_jsonl_component_path = self._compute_component_path_of_pages_config_value(device.path, device.config)
        if pages_jsonl_component_path is not None:
            pages_jsonl_component = self._create_text_component_from_path(
                device_cfg_dir_root=device.path,
                path=pages_jsonl_component_path,
            )
            referenced_jsonl_components.add(pages_jsonl_component)

        result.extend(referenced_jsonl_components)
        result.extend(referenced_cmd_components)
        result.extend(referenced_image_components)
        return result

    def _find_referenced_cmd_components(
        self, cmd_components: List[CmdComponent]
    ) -> Set[CmdComponent]:
        result = set()

        system_components = list(filter(lambda x: x.name in SYSTEM_SCRIPTS, cmd_components))

        # TODO: this should also consider the hierarchy, if a cmd component is not a system component, and
        #  it is also not referenced anywhere, the component is not relevant

        for component in cmd_components:
            cmd_references = self._find_cmd_references_in_cmd_component(component)
            for match in cmd_references:
                matching_components = list(filter(lambda x: x.name == match, cmd_components))
                if len(matching_components) <= 0:
                    found_component_names = ','.join([c.name for c in cmd_components])
                    raise AssertionError(
                        f"Referenced CMD component not found: {match}, only found: {found_component_names}")
                result.update(matching_components)

        # system components should always be included
        result.update(system_components)

        return result

    @staticmethod
    def _find_cmd_references_in_cmd_component(component: TextComponent) -> Set[str]:
        result = set()
        pattern = re.compile(r"L:/(.*\.cmd)")
        for line in component.content.splitlines():
            matches = re.findall(pattern, line)
            result.update(matches)
        return result

    def _find_referenced_jsonl_components(
        self,
        cmd_components: List[CmdComponent],
        jsonl_components: List[JsonlComponent]
    ) -> Set[JsonlComponent]:
        referenced_jsonl_components = set()
        for component in cmd_components:
            jsonl_references = self._find_jsonl_references_in_cmd_component(component)
            for match in jsonl_references:
                matching_components = list(filter(lambda x: x.name == match, jsonl_components))
                if len(matching_components) <= 0:
                    found_component_names = ','.join([c.name for c in jsonl_components])
                    raise AssertionError(
                        f"Referenced JSONL component not found: {match}, only found: {found_component_names}")
                referenced_jsonl_components.update(matching_components)

        return referenced_jsonl_components

    @staticmethod
    def _find_jsonl_references_in_cmd_component(component: TextComponent) -> Set[str]:
        result = set()
        pattern = re.compile(r"L:/(.*\.jsonl)")
        for line in component.content.splitlines():
            matches = re.findall(pattern, line)
            result.update(matches)
        return result

    def _find_referenced_image_components(
        self, cmd_components: List[CmdComponent],
        jsonl_components: List[JsonlComponent],
        image_components: List[ImageComponent]
    ) -> Set[ImageComponent]:
        """
        Find all image components that are referenced in the given cmd and jsonl components.
        :param cmd_components:
        :param jsonl_components:
        :param image_components:
        :return: set of referenced image components
        """
        referenced_image_components = set()
        for component in cmd_components:
            image_references = self._find_image_references_in_cmd_component(component)
            for match in image_references:
                matching_components = list(filter(lambda x: x.name == match, image_components))
                if len(matching_components) <= 0:
                    found_component_names = ','.join([c.name for c in image_components])
                    raise AssertionError(
                        f"Referenced image component not found: {match}, only found: {found_component_names}")
                referenced_image_components.update(matching_components)

        for component in jsonl_components:
            image_references = self._find_image_references_in_jsonl_component(component)
            for match in image_references:
                matching_components = list(filter(lambda x: x.name == match, image_components))
                if len(matching_components) <= 0:
                    found_component_names = ','.join([c.name for c in image_components])
                    raise AssertionError(
                        f"Referenced image component not found: {match}, only found: {found_component_names}")
                referenced_image_components.update(matching_components)

        return referenced_image_components

    @staticmethod
    def _find_image_references_in_cmd_component(component: TextComponent) -> Set[str]:
        result = set()
        pattern = re.compile(r"L:/(.*\.(?:png|bin))")
        for line in component.content.splitlines():
            matches = re.findall(pattern, line)
            result.update(matches)
        return result

    @staticmethod
    def _find_image_references_in_jsonl_component(component: TextComponent) -> Set[str]:
        result = set()
        pattern = re.compile(r"L:/(.*\.(?:png|bin))")
        for line in component.content.splitlines():
            matches = re.findall(pattern, line)
            result.update(matches)
        return result

    def determine_device_jsonl_component_order_for_cmd(
        self, device: Device, cmd_component: CmdComponent) -> List[JsonlComponent]:
        """
        Determines the order of the jsonl components based on the order of their reference in cmd components.

        :param device: The device to analyze.
        :param cmd_component: The cmd component to analyze.
        :return: The ordered list of jsonl components.
        """
        device_cmd_components = device.cmd
        device_jsonl_components = device.jsonl

        return self._determine_device_jsonl_component_order_for_cmd(
            cmd_component=cmd_component,
            device_cmd_components=device_cmd_components,
            device_jsonl_components=device_jsonl_components
        )

    def _determine_device_jsonl_component_order_for_cmd(
        self, cmd_component: CmdComponent,
        device_cmd_components: List[CmdComponent],
        device_jsonl_components: List[JsonlComponent]) -> List[JsonlComponent]:
        """
        Determines an ordered list of all referenced jsonl components based on their reference in the given cmd component.

        :param cmd_component: The cmd component to analyze.
        :param device_cmd_components: The list of cmd components to check for references.
        :param device_jsonl_components: The list of jsonl components to order.
        :return: The ordered list of jsonl components.
        """
        jsonl_components = []

        for cmd in cmd_component.commands:
            if cmd.endswith(".jsonl"):
                jsonl_component_name = cmd.split("/")[-1]
                jsonl_component = next(
                    filter(lambda x: x.name == jsonl_component_name, device_jsonl_components), None
                )
                if jsonl_component is None:
                    raise AssertionError(f"Component {jsonl_component_name} not found in device jsonl components")
                jsonl_components.append(jsonl_component)
            elif cmd.endswith(".cmd"):
                cmd_component_name = cmd.split("/")[-1]
                cmd_component = next(
                    filter(lambda x: x.name == cmd_component_name, device_cmd_components), None
                )
                if cmd_component is None:
                    raise AssertionError(f"Component {cmd_component_name} not found in device cmd components")
                jsonl_components.extend(
                    self._determine_device_jsonl_component_order_for_cmd(
                        cmd_component,
                        device_cmd_components,
                        device_jsonl_components
                    )
                )

        return jsonl_components

    @staticmethod
    def _write_output(device: Device, component: Component, output_content: str | bytes):
        device.output_dir.mkdir(parents=True, exist_ok=True)

        component_output_file = Path(
            device.output_dir,
            component.name
        )

        if isinstance(output_content, bytes):
            component_output_file.write_bytes(output_content)
        elif isinstance(output_content, str):
            component_output_file.write_text(output_content)
        else:
            raise AssertionError(f"Unsupported output type: {type(output_content)}")

    @staticmethod
    def _clear_output(device: Device):
        if device.output_dir.exists():
            shutil.rmtree(device.output_dir)



import json
from pathlib import Path
from typing import Dict

from openhasp_config_manager.cli.common import _create_config_manager, _analyze_and_filter
from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
from openhasp_config_manager.ui.util import success, error


async def c_state(config_dir: Path, device: str, object: str, state: str):
    try:
        state_json = json.loads(state)
        if not isinstance(state_json, Dict):
            raise Exception("State must be a JSON object.")

        config_manager = _create_config_manager(config_dir, Path("./nonexistent"))
        filtered_devices, ignored_devices = _analyze_and_filter(config_manager=config_manager, device_filter=device)
        if len(filtered_devices) <= 0:
            raise Exception(f"No device matches the filter: {device}")
        for device in filtered_devices:
            client = OpenHaspClient(device)
            await client.set_object_properties(object, state_json)
        success("Done!")
    except Exception as ex:
        error(str(ex))



from typing import Tuple

from PIL import Image
from temppathlib import NamedTemporaryFile


class OpenHaspImageProcessor:

    def image_to_rgb565(self, in_image, out_image, size: Tuple[int or None, int or None], fitscreen: bool):
        """
        Transforms an image to rgb565 format according to LVGL requirements.
        :param in_image: the image to transform
        :param out_image: the output file
        :param size: the size of the output image
        :param fitscreen: if True, the image will be resized to fit the screen
        """
        import struct

        if in_image.startswith("http"):
            tmp_file_container = self._fetch_image_from_url(in_image)
        else:
            # copy file to temp file
            tmp_file_container = NamedTemporaryFile(delete=False)
            import shutil
            shutil.copy(in_image, tmp_file_container.path)

        try:
            im = Image.open(tmp_file_container.file)
            original_width, original_height = im.size
            width, height = size

            if not fitscreen:
                width = min(w for w in [width, original_width] if w is not None and w > 0)
                height = min(h for h in [height, original_height] if h is not None and h > 0)
                im.thumbnail((height, width), Image.Resampling.LANCZOS)
            else:
                im = im.resize((height, width), Image.Resampling.LANCZOS)
            width, height = im.size  # actual size after resize

            out_image.write(struct.pack("I", height << 21 | width << 10 | 4))

            img = im.convert("RGB")
            for pix in img.getdata():
                r = (pix[0] >> 3) & 0x1F
                g = (pix[1] >> 2) & 0x3F
                b = (pix[2] >> 3) & 0x1F
                out_image.write(struct.pack("H", (r << 11) | (g << 5) | b))
            out_image.flush()

            im.close()
            img.close()
        finally:
            tmp_file_container.path.unlink()

    @staticmethod
    def _fetch_image_from_url(in_image) -> NamedTemporaryFile:
        import requests

        # consider last part of the url as filename
        filename = in_image.split("/")[-1]
        # remove url args, if any
        filename = filename.split("?")[0]
        response = requests.get(in_image, stream=True)
        response.raise_for_status()
        content_type = response.headers.get('content-type', None)
        content = response.content

        # add file extension based on content type
        from mimetypes import guess_extension
        guess = guess_extension(content_type)
        if guess is not None:
            filename = filename + guess

        tmp_image = NamedTemporaryFile(suffix=filename, delete=False)
        tmp_image.file.write(content)
        return tmp_image



import logging
from typing import List, Union

import click

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


def error(text: str):
    prefix = click.style(" FAIL ", fg="black", bg="red")
    click.echo(f"{prefix} ", nl=False)
    echo(text)


def warn(text: str):
    prefix = click.style(" WARN ", fg="black", bg="yellow")
    click.echo(f"{prefix} ", nl=False)
    echo(text)


def info(text: str):
    prefix = click.style(" INFO ", fg="white", bg=(33, 33, 33))
    click.echo(f"{prefix} ", nl=False)
    echo(text)


def success(text: str):
    prefix = click.style("  OK  ", fg="black", bg="green")
    click.echo(f"{prefix} ", nl=False)
    echo(text)


def echo(text: Union[str, List] = "", fg_color=None, bg_color=None):
    """
    Prints a text to the console
    :param text: the text
    :param fg_color: an optional foreground (text) color
    :param bg_color: an optional background color
    """
    if text is not click.termui and text is not str:
        text = str(text)
    if fg_color:
        text = click.style(text, fg=fg_color, bg=bg_color)
    if len(text) > 0:
        LOGGER.debug(text)
    click.echo(text)


def print_diff_to_console(diff_output):
    for line in diff_output.splitlines():
        if line.startswith("+"):
            echo(line, fg_color="green")
        elif line.startswith("-"):
            echo(line, fg_color="red")
        else:
            echo(line)



import abc


class Validator(metaclass=abc.ABCMeta):
    """
    Generic interface to implement validators used when validating the configuration.
    """

    @abc.abstractmethod
    def validate(self, data: str):
        """
        Validates the given input data.

        If the input is considered valid, the method will return.

        If the input is considered invalid, an error will be raised further
        detailing the reasons why it is considered invalid.

        :param data: the input data to analyze
        """
        raise NotImplementedError



from dataclasses import dataclass

from openhasp_config_manager.openhasp_client.model.configuration.device_config import DeviceConfig


@dataclass
class OpenhaspConfigManagerConfig:
    device: DeviceConfig



import logging
import re
from typing import Dict

from openhasp_config_manager.openhasp_client.model.configuration.config import Config
from openhasp_config_manager.processing.jsonl import JsonlObjectProcessor

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class ObjectDimensionsProcessor(JsonlObjectProcessor):
    """
    Used to process .jsonl files to add support for additional features.
    """
    PERCENTAGE_REGEX_PATTERN = re.compile(r"^\d+(\.\d+)?%$")

    def process(
        self, input: Dict, config: Config, template_vars: Dict[str, any]
    ) -> Dict:
        result: Dict[str, any] = {}
        for key, value in input.items():
            if isinstance(value, str) and re.match("[xywh]", key) and re.match(self.PERCENTAGE_REGEX_PATTERN, value):
                numeric_value = self._parse_percentage(value)

                total_width = config.openhasp_config_manager.device.screen.width
                total_height = config.openhasp_config_manager.device.screen.height

                if key in ["x", "w"]:
                    total = total_width
                else:
                    total = total_height

                result[key] = self._percentage_of(numeric_value, total)
            else:
                result[key] = value

        # normalize value types, in case of templates
        for key, value in result.items():
            try:
                match = re.match(
                    "^(page|id|x|y|w|h|text_font|value_font|radius|pad_.+|margin_.+|border_width|min|max|prev|next).*$",
                    string=key,
                )

                if match and isinstance(value, str):
                    result[key] = int(float(value))
            except Exception as ex:
                LOGGER.exception(ex)
                print(f"{key}: {value}: {ex}, {input}")

        return result

    @staticmethod
    def _percentage_of(percentage: float, total: int) -> int:
        """
        :param percentage: 0..100
        :param total: value for a percentage of 100
        :return: value according to input
        """
        return int((percentage * total) / 100)

    @staticmethod
    def _parse_percentage(value: str) -> float:
        return float(str(value).replace('%', ''))


class ObjectThemeProcessor(JsonlObjectProcessor):
    """
    Used to apply default theming from "theme.obj.*" values
    """

    def process(
        self, input: Dict, config: Config, template_vars: Dict[str, any]
    ) -> Dict:
        obj_key = input.get("obj", None)
        if obj_key is None:
            return input

        theme_values = template_vars.get("theme", {}).get("obj", {}).get(obj_key, {})
        for key, value in theme_values.items():
            if key not in input:
                input[key] = value

        return input



def clear_layout(layout):
    """Recursively delete all widgets and layouts in the given layout."""
    if layout is not None:
        while layout.count():
            child = layout.takeAt(0)
            if child.widget() is not None:
                child.widget().setParent(None)
                child.widget().deleteLater()
            elif child.layout() is not None:
                clear_layout(child.layout())



from collections import OrderedDict
from typing import List, Tuple, Dict, Set

import qtawesome as qta
from PyQt6 import QtCore
from PyQt6.QtCore import QSize, Qt, QRect, QRectF
from PyQt6.QtGui import QPainter, QBrush, QColor, QMouseEvent, QPainterPath, QFont, QPen
from PyQt6.QtWidgets import QWidget, QVBoxLayout, QSizePolicy, QPushButton, QHBoxLayout, QTextEdit
from orjson import orjson

from openhasp_config_manager.manager import ConfigManager
from openhasp_config_manager.openhasp_client.model.component import JsonlComponent
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.openhasp_client.openhasp import IntegratedIcon
from openhasp_config_manager.ui.qt.util import clear_layout


class OpenHaspDevicePagesData:
    def __init__(self, device: Device, name: str, jsonl_components: List[JsonlComponent]):
        self.device = device
        self.name = name
        self.jsonl_components = jsonl_components


class PageLayoutEditorWidget(QWidget):
    def __init__(self, config_manager: ConfigManager):
        super().__init__()
        self.config_manager = config_manager
        self.device_processor = None
        # map of <json component name, list of objects>
        self.jsonl_component_objects: OrderedDict[str, List[Dict]] = OrderedDict()

        self.device_pages_data = None
        self.current_index = 1

        self.create_layout()

    def create_layout(self):
        self.layout = QVBoxLayout(self)

        self.page_selector = self._create_page_selector()
        self.layout.addWidget(self.page_selector)

        self.preview_container = QWidget()
        self.preview_container.setLayout(QVBoxLayout())
        self.layout.addWidget(self.preview_container)

    def _on_previous_page_clicked(self):
        self.previous_page_index()

    def _on_next_page_clicked(self):
        self.next_page_index()

    def clear(self):
        clear_layout(self.preview_container.layout())
        self.current_index = 1
        self.jsonl_component_objects.clear()

    def set_data(self, device_pages_data: OpenHaspDevicePagesData or None):
        self.device_pages_data = device_pages_data
        self.clear()
        if device_pages_data is None:
            return
        self.device_processor = self.config_manager.create_device_processor(device_pages_data.device)

        # load jsonl component objects
        for jsonl_component in self.device_pages_data.jsonl_components:
            normalized_jsonl_component = self.device_processor.normalize(self.device_pages_data.device, jsonl_component)
            objects_in_jsonl = normalized_jsonl_component.splitlines()
            loaded_objects = list(map(orjson.loads, objects_in_jsonl))
            self.jsonl_component_objects[jsonl_component.name] = loaded_objects

        self.page_preview_widget = PagePreviewWidget(device_pages_data, [])
        self.preview_container.layout().addWidget(self.page_preview_widget)

        self.page_jsonl_preview = PageJsonlPreviewWidget(device_pages_data)
        self.preview_container.layout().addWidget(self.page_jsonl_preview)

        self.set_page_index(index=1)

    def set_page_index(self, index: int):
        print("Setting page index", index)
        self.current_index = index
        self.page_objects = self.get_page_objects(index=index)
        print(f"Page objects: {self.page_objects}")
        self.page_preview_widget.set_objects(self.page_objects)

        self.page_jsonl_preview.set_objects(self.page_objects)

    def next_page_index(self):
        """
        Cycle to the next available page index, skipping page 0 and rolling over to the first page.
        """
        usable_page_indices = self.get_used_page_indices() - {0}
        usable_page_indices = list(sorted(usable_page_indices))
        current_page_index_position_in_set = usable_page_indices.index(self.current_index)
        new_page_index_position_in_set = (current_page_index_position_in_set + 1) % len(usable_page_indices)
        self.set_page_index(usable_page_indices[new_page_index_position_in_set])

    def previous_page_index(self):
        """
        Cycle to the previous available page index, skipping page 0 and rolling over to the last page.
        """
        usable_page_indices = self.get_used_page_indices() - {0}
        usable_page_indices = list(sorted(usable_page_indices))
        current_page_index_position_in_set = usable_page_indices.index(self.current_index)
        new_page_index_position_in_set = (current_page_index_position_in_set - 1) % len(usable_page_indices)
        self.set_page_index(usable_page_indices[new_page_index_position_in_set])

    def get_used_page_indices(self) -> Set[int]:
        """
        Get the used page indices from the jsonl component objects.
        :return: a list of used page indices
        """
        if self.device_pages_data is None:
            return set()

        used_page_indices = set()
        for jsonl_component_name, objects_in_jsonl in self.jsonl_component_objects.items():
            for obj in objects_in_jsonl:
                object_page_index = obj.get("page", None)
                if object_page_index is not None:
                    used_page_indices.add(object_page_index)

        print(f"Used page indices: {used_page_indices}")

        return used_page_indices

    def get_page_objects(self, index: int, include_global: bool = True) -> List[dict]:
        """
        Get the page objects for the given page index.
        :param index: the index of the page
        :return: a list of objects for the page
        """
        if self.device_pages_data is None:
            return []

        result = []
        for jsonl_component_name, objects_in_jsonl in self.jsonl_component_objects.items():
            result = result + objects_in_jsonl

        # Filter the objects based on the index
        result = [
            obj for obj in result if
            obj.get("page") == index or (obj.get("page") == 0 if include_global else False)
        ]

        # Filter hidden objects
        result = [obj for obj in result if not obj.get("hidden", False)]

        # Draw objects on page 0 last (on top)
        result.sort(key=lambda obj: obj.get("page") == 0)

        return result

    def _create_page_selector(self) -> QWidget:
        page_selector_widget = QWidget()
        page_selector_widget.setLayout(QHBoxLayout())

        self._previous_page_button_widget = QPushButton("Previous Page")
        self._previous_page_button_widget.clicked.connect(self._on_previous_page_clicked)
        page_selector_widget.layout().addWidget(self._previous_page_button_widget)

        self._next_page_button_widget = QPushButton("Next Page")
        self._next_page_button_widget.clicked.connect(self._on_next_page_clicked)
        page_selector_widget.layout().addWidget(self._next_page_button_widget)

        return page_selector_widget


class PageJsonlPreviewWidget(QTextEdit):
    def __init__(self, page: OpenHaspDevicePagesData, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setReadOnly(True)
        self.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        self.setFont(QFont("Roboto Mono", 10))

        self.set_page(page)

        size_policy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
        self.setSizePolicy(size_policy)

    def set_page(self, page: OpenHaspDevicePagesData):
        self.page = page
        self.setText(page.jsonl_components[0].content)

    def set_objects(self, page_objects: List[dict]):
        """
        Set the objects for the page.
        :param page_objects: the list of objects to set
        """
        sorted_page_objects = sorted(page_objects, key=lambda obj: (
        obj.get("page", 0), obj.get("id", 0), obj.get("y", ""), obj.get("x", "")))

        content = "\n".join(map(lambda x: orjson.dumps(x).decode(), sorted_page_objects))
        self.setText(content)


class PagePreviewWidget(QWidget):
    clickedValue = QtCore.pyqtSignal(int)

    def __init__(self, page: OpenHaspDevicePagesData, page_objects: List[dict], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.page = page
        self.objects: List[dict] = page_objects

        self._padding = 0

        size_policy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        size_policy.setHeightForWidth(True)
        self.setSizePolicy(size_policy)

    @property
    def page_width(self) -> int:
        return self.page.device.config.openhasp_config_manager.device.screen.width

    @property
    def page_height(self) -> int:
        return self.page.device.config.openhasp_config_manager.device.screen.height

    def sizeHint(self):
        return QSize(
            self.page_width,
            self.page_height,
        )

    def heightForWidth(self, width):
        # ratio = self.page_width / self.page_height
        return int((width / self.page_width) * self.page_height)

    def set_objects(self, loaded_objects: List[dict]):
        self.objects = loaded_objects
        self._trigger_refresh()

    def _calculate_clicked_value(self, e: QMouseEvent):
        parent = self.parent()
        vmin, vmax = parent.minimumHeight(), parent.maximumHeight()
        d_height = self.size().height() + (self._padding * 2)
        click_y = e.pos().y() - self._padding

        pc = (d_height - click_y) / d_height
        value = vmin + pc * (vmax - vmin)
        self.clickedValue.emit(value)

    def mouseMoveEvent(self, e: QMouseEvent):
        pass
        # self._calculate_clicked_value(e)

    def mousePressEvent(self, e: QMouseEvent):
        self._calculate_clicked_value(e)

    def paintEvent(self, e):
        painter = QPainter(self)

        brush = QBrush()
        brush.setColor(QColor('black'))
        brush.setStyle(Qt.BrushStyle.SolidPattern)
        rect = QRect(0, 0, painter.device().width(), painter.device().height())
        painter.eraseRect(rect)
        painter.fillRect(rect, brush)

        padding = 0

        # Define our canvas.
        d_height = painter.device().height() - (padding * 2)
        d_width = painter.device().width() - (padding * 2)

        # Draw the objects
        for i, obj in enumerate(self.objects):
            object_type = obj.get("obj", None)
            if object_type is None:
                continue
            elif object_type == "btn":
                self._draw_button(painter, obj, padding, d_width, d_height)
            elif object_type == "label":
                self._draw_label(painter, obj, padding, d_width, d_height)
            elif object_type == "slider":
                self._draw_slider(painter, obj, padding, d_width, d_height)
            elif object_type == "switch":
                self._draw_switch(painter, obj, padding, d_width, d_height)
            elif object_type == "img":
                self._draw_image(painter, obj, padding, d_width, d_height)
            elif object_type == "bar":
                self._draw_bar(painter, obj, padding, d_width, d_height)
            elif object_type == "msgbox":
                self._draw_messagebox(painter, obj, padding, d_width, d_height)
            elif object_type == "obj":
                self._draw_obj(painter, obj, padding, d_width, d_height)
            else:
                print(f"Unknown object type: {object_type}")

    def _trigger_refresh(self):
        self.update()

    def _draw_label(self, painter, obj, padding, d_width, d_height):
        """
        Draws a label on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing label with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)

        text = obj.get("text", "")
        text_color = obj.get("text_color", None)
        text_font = obj.get("text_font", 25)
        text_align = obj.get("align", "left")
        object_bg_color = obj.get("bg_color", None)
        border_width = obj.get("border_width", 0)
        border_color = obj.get("border_color", None)

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color,
            border_width=border_width, border_color=border_color,
        )

        text_alignment_flag = Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft
        if text_align == "center":
            text_alignment_flag = Qt.AlignmentFlag.AlignCenter
        elif text_align == "right":
            text_alignment_flag = Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight
        self._draw_scaled_text(
            painter, x, y, width, height, padding, d_width, d_height,
            text=text, text_color=text_color, pixel_size=text_font,
            flags=text_alignment_flag,
            border_width=border_width, border_color=border_color,
        )

    def _draw_button(self, painter, obj, padding, d_width, d_height):
        """
        Draws a button on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing button with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        object_bg_color = obj.get("bg_color", "blue")
        text = obj.get("text", "")
        text_font = obj.get("text_font", 25)
        text_color = obj.get("text_color", None)
        radius = obj.get("radius", 0)
        action = obj.get("action", None)

        brush = QBrush()
        brush.setColor(QColor(object_bg_color))
        brush.setStyle(Qt.BrushStyle.SolidPattern)

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color, corner_radius=radius,
        )

        self._draw_scaled_text(
            painter, x, y, width, height, padding, d_width, d_height,
            text=text, text_color=text_color, pixel_size=text_font
        )

    def _draw_slider(self, painter, obj, padding, d_width, d_height):
        """
        Draws a slider on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing slider with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        object_bg_color = obj.get("bg_color", "gray")
        knob_color = obj.get("bg_color20", "lightgray")
        radius = obj.get("radius", 0)
        knob_radius = obj.get("radius20", 0)
        slider_min = obj.get("min", 0)
        slider_max = obj.get("max", 100)
        slider_value = obj.get("value", 0)

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color, corner_radius=radius,
        )

        # Draw the slider value
        self._draw_scaled_circle(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=knob_color, radius=knob_radius,
        )

    def _draw_switch(self, painter, obj, padding, d_width, d_height):
        """
        Draws a switch on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing switch with id {object_id}: {obj}")

        value = obj.get("val", 0)
        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        corner_radius = obj.get("radius00", 0)
        object_bg_color = obj.get("bg_color", "green")
        text_font = obj.get("text_font", 25)
        text_color = obj.get("text_color", None)

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color, corner_radius=corner_radius,
        )

        # Draw the knob
        knob_radius = obj.get("radius20", 0)
        knob_color = obj.get("bg_color20", "lightgray")

        knob_x = x + (width - 1) if value else x
        self._draw_scaled_circle(
            painter, knob_x, y, width, height, padding, d_width, d_height,
            fill_color=knob_color, radius=knob_radius
        )

        # Draw the text
        self._draw_scaled_text(
            painter, x, y, width, height, padding, d_width, d_height,
            text=obj.get("text", ""), text_color=text_color, pixel_size=text_font
        )

    def _draw_image(self, painter, obj, padding, d_width, d_height):
        """
        Draws an image on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing image with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        object_bg_color = obj.get("bg_color", "yellow")

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color
        )

    def _draw_bar(self, painter, obj, padding, d_width, d_height):
        """
        Draws a bar on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing bar with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        corner_radius = obj.get("radius", height)
        text_font = obj.get("text_font", 25)
        text_color = obj.get("text_color", None)
        object_bg_color = obj.get("bg_color", "purple")

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color, corner_radius=corner_radius
        )

        # Draw the text
        self._draw_scaled_text(
            painter, x, y, width, height, padding, d_width, d_height,
            text=obj.get("text", ""), text_color=text_color, pixel_size=text_font
        )

    def _draw_messagebox(self, painter, obj, padding, d_width, d_height):
        """
        Draws a messagebox on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing messagebox with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        text_font = obj.get("text_font", 25)
        text_color = obj.get("text_color", None)
        object_bg_color = obj.get("bg_color", "orange")

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color
        )

        # Draw the text
        self._draw_scaled_text(
            painter, x, y, width, height, padding, d_width, d_height,
            text=obj.get("text", ""), text_color=text_color, pixel_size=text_font
        )

    def _draw_obj(self, painter, obj, padding, d_width, d_height):
        """
        Draws a generic object on the canvas.
        :param painter: the painter to use for drawing
        :param obj: the object to draw
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        """
        object_id = obj.get("id", None)
        print(f"Drawing object with id {object_id}: {obj}")

        x = obj.get("x", 0)
        y = obj.get("y", 0)
        width = obj.get("w", 50)
        height = obj.get("h", 50)
        object_bg_color = obj.get("bg_color", "gray")

        self._draw_scaled_rect(
            painter, x, y, width, height, padding, d_width, d_height,
            fill_color=object_bg_color
        )

    def _draw_scaled_rect(
        self,
        painter, x, y, width, height, padding, d_width, d_height,
        fill_color: str, corner_radius: int = 0,
        border_width: int = 0, border_color: str = "#FFFFFF"
    ):
        """
        Helper method to draw a scaled rectangle on the canvas.
        :param painter: the painter to use for drawing
        :param x: the x position of the rectangle
        :param y: the y position of the rectangle
        :param width: the width of the rectangle
        :param height: the height of the rectangle
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        :param fill_color: the fill color of the rectangle
        :param corner_radius: the corner radius of the rectangle
        :param border_width: the border width of the rectangle
        :param border_color: the border color of the rectangle
        """
        if fill_color is None:
            return
        if border_color is None:
            border_color = "#FFFFFF"
        if corner_radius is None:
            corner_radius = 0

        scaled_x, scaled_y, scaled_width, scaled_height = self.__get_scaled_rect(
            x, y, width, height, padding, d_width, d_height
        )

        scale_factor = 0.8 * (d_height / self.page_height)
        scaled_corner_radius = min(scaled_height / 2, int(corner_radius * scale_factor))

        rect = QRectF(
            padding + scaled_x,
            padding + scaled_y,
            scaled_width,
            scaled_height
        )
        brush = QBrush(QColor(fill_color))
        brush.setStyle(Qt.BrushStyle.SolidPattern)

        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        path = QPainterPath()
        path.addRoundedRect(rect, scaled_corner_radius, scaled_corner_radius)
        painter.fillPath(path, brush)

        # draw border
        if border_width > 0:
            scaled_border_width = int(border_width * scale_factor)
            border_pen = QPen(QColor(border_color))
            border_pen.setWidth(scaled_border_width)
            path = QPainterPath()

            # inset rect to account for border width
            border_rect = QRectF(
                padding + scaled_x + scaled_border_width,
                padding + scaled_y + scaled_border_width,
                scaled_width - (scaled_border_width * 2),
                scaled_height - (scaled_border_width * 2)
            )
            path.addRoundedRect(border_rect, scaled_corner_radius, scaled_corner_radius)
            painter.strokePath(path, border_pen)

    def _draw_scaled_circle(self, painter, x, y, width, height, padding, d_width, d_height, fill_color, radius):
        """
        Helper method to draw a scaled circle on the canvas.
        :param painter: the painter to use for drawing
        :param x: the x position of the circle
        :param y: the y position of the circle
        :param width: the width of the circle
        :param height: the height of the circle
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        :param fill_color: the fill color of the circle
        :param radius: the radius of the circle
        """
        if fill_color is None:
            return

        if radius <= 0:
            return

        scaled_x, scaled_y, scaled_width, scaled_height = self.__get_scaled_rect(
            x, y, width, height, padding, d_width, d_height
        )

        scale_factor = 0.8 * (d_height / self.page_height)
        scaled_radius = int(radius * scale_factor)
        scaled_radius = min(scaled_radius, scaled_height // 2)

        rect = QRectF(
            padding + scaled_x,
            padding + scaled_y,
            scaled_radius * 2,
            scaled_radius * 2
        )
        brush = QBrush(QColor(fill_color))
        brush.setStyle(Qt.BrushStyle.SolidPattern)

        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        path = QPainterPath()
        path.addEllipse(rect)
        painter.fillPath(path, brush)

    def _draw_scaled_text(
        self, painter, x, y, width, height, padding, d_width, d_height, text: str = "",
        text_color: str = "white", pixel_size: int = 48, flags: int = Qt.AlignmentFlag.AlignCenter,
        border_width: int = 0, border_color: str = None
    ):
        """
        Helper method to draw scaled text on the canvas.
        :param painter: the painter to use for drawing
        :param x: the x position of the text
        :param y: the y position of the text
        :param width: the width of the text
        :param height: the height of the text
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        :param text: the text to draw
        :param text_color: the color of the text
        :param pixel_size: the pixel size of the text
        :param flags: the text alignment flags
        :param border_width: the border width of the text
        :param border_color: the border color of the text
        """
        if not text:
            return
        if text_color is None:
            text_color = "white"

        scaled_x, scaled_y, scaled_width, scaled_height = self.__get_scaled_rect(
            x, y, width, height, padding, d_width, d_height
        )

        scale_factor = 0.7 * (d_height / self.page_height)
        scaled_pixel_size = int(pixel_size * scale_factor)

        # draw border first
        if border_width > 0:
            scaled_border_width = int(border_width * scale_factor)
            border_pen = QPen(QColor(border_color))
            border_pen.setWidth(scaled_border_width)
            path = QPainterPath()

            # inset rect to account for border width
            border_rect = QRectF(
                padding + scaled_x + scaled_border_width,
                padding + scaled_y + scaled_border_width,
                scaled_width - (scaled_border_width * 2),
                scaled_height - (scaled_border_width * 2)
            )

            path.addRect(border_rect)
            painter.strokePath(path, border_pen)

        # replace unicode chars with icons from material design icons
        text = self._replace_unicode_with_icons(text)

        rect = QRectF(
            padding + scaled_x,
            padding + scaled_y,
            scaled_width,
            scaled_height
        )
        painter.setPen(QColor(text_color))
        font = QFont("Roboto Condensed", scaled_pixel_size, QFont.Weight.Normal)
        # print(font.family())
        # print(font.exactMatch())
        painter.setFont(font)
        painter.drawText(rect, flags, text)

    def __get_scaled_rect(self, x, y, width, height, padding, d_width, d_height) -> Tuple[int, int, int, int]:
        """
        Helper method to get the scaled rectangle for the given parameters.
        :param x: the x position of the rectangle
        :param y: the y position of the rectangle
        :param width: the width of the rectangle
        :param height: the height of the rectangle
        :param padding: the padding around the canvas
        :param d_width: the width of the canvas
        :param d_height: the height of the canvas
        :return: a tuple of (x, y, width, height) for the scaled rectangle
        """
        scaled_x = int((x / self.page_width) * d_width)
        scaled_y = int((y / self.page_height) * d_height)
        scaled_width = int((width / self.page_width) * d_width)
        scaled_height = int((height / self.page_height) * d_height)

        return scaled_x, scaled_y, scaled_width, scaled_height

    def _replace_unicode_with_icons(self, text: str) -> str:
        """
        Replaces unicode characters in the text with icons from material design icons.
        :param text: the text to process
        :return: the text with unicode characters replaced with icons
        """
        for unicode_char, icon_name in IntegratedIcon.entries():
            icon_charmap = qta.charmap(f"mdi6.{icon_name}")
            text = text.replace(unicode_char, icon_charmap)

        return text



import logging
import re
from typing import Dict, List

import jinja2
from jinja2 import BaseLoader

from openhasp_config_manager.ui.util import echo, error

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


def render_dict_recursive(
        input: Dict,
        template_vars: Dict,
        result_key_path: List[str] = None,
) -> Dict[str, any]:
    """
    Recursively called function to resolve templates within the given input dict
    :param input: the dict which (possibly) contains templates
    :param template_vars: a map specifying the value of template variables # TODO: this might need context as well... right?
    :param result_key_path:
    :return:
    """
    if result_key_path is None:
        result_key_path = []

    result = {}

    finished = False
    progress = 0
    last_pending = None
    while not finished:
        pending = []
        finished = True
        keys = list(input.keys())
        for key in keys:
            value = input[key]
            # key
            rendered_key = key
            key_undefined = False
            if "{{" in key:
                try:
                    rendered_key = _render_template(key, template_vars)
                    key_undefined = _has_undeclared_variables(rendered_key)
                except Exception as ex:
                    error(f"Undefined key: {key_undefined}, value: {key}")
                    key_undefined = True

            # value
            value_undefined = False
            rendered_value = value
            if isinstance(value, dict) and rendered_key is not None and not key_undefined:
                rendered_value = render_dict_recursive(
                    input=value,
                    template_vars=template_vars,
                    result_key_path=result_key_path + [rendered_key]
                )
            elif isinstance(value, list):
                try:
                    rendered_value = list(map(lambda x: _render_template(x, template_vars), value))
                    value_undefined = any(map(lambda x: _has_undeclared_variables(x), rendered_value))
                except Exception as ex:
                    # print(f"Undefined value: {value_undefined}, value: {value}")
                    value_undefined = True
            elif isinstance(value, str):
                try:
                    rendered_value = _render_template(value, template_vars)
                    value_undefined = _has_undeclared_variables(rendered_value)
                except Exception as ex:
                    # print(f"Undefined value: {value_undefined}, value: {value}")
                    value_undefined = True

            if key_undefined:
                pending.append(key)
            elif key in pending:
                pending.remove(key)
            if value_undefined:
                pending.append(f"{key}__value")
            elif f"{key}__value" in pending:
                pending.remove(f"{key}__value")

            if key_undefined or value_undefined:
                # still has undefined keys
                # print(f"Undefined keys: {key_undefined}, Undefined values: {value_undefined}")
                finished = False
            else:
                # add it to the template_vars data at the correct location
                # which allows later iterations to render the template correctly
                tmp = template_vars
                for idx, p in enumerate(result_key_path):
                    if p not in tmp.keys():
                        tmp[p] = {}
                    tmp = tmp[p]
                tmp.pop(key, None)
                tmp[rendered_key] = rendered_value

                if rendered_key not in result or result[rendered_key] != rendered_value:
                    progress += 1

                # create the resulting (sub-)dictionary
                result[rendered_key] = rendered_value

                finished = finished and True

        if len(pending) <= 0:
            return result
        elif pending == last_pending:
            echo(f"No progression while rendering templates, unable to render: {pending}")
            return result
        else:
            last_pending = pending

    return result


_j2_env = jinja2.Environment(loader=BaseLoader(), undefined=jinja2.DebugUndefined)
_template_cache = {}


def _render_template(content: str, template_vars: Dict[str, str]) -> str:
    inner_templates = re.findall(r"\{\{.+}}", content[2:-2])
    for inner_template in inner_templates:
        if inner_template != content[2:-2]:
            rendered = _render_template(inner_template, template_vars)
            content = content.replace(inner_template, rendered)
    try:
        if content not in _template_cache:
            _template_cache[content] = _j2_env.from_string(content)
        template = _template_cache[content]
        rendered = template.render(template_vars)
        return rendered
    except Exception as ex:
        LOGGER.exception(ex)
        print(f"{ex}: \n\n{content}\n\n {template_vars}")
        raise ex


_template_ast_cache = {}


def _has_undeclared_variables(rendered_value: str):
    from jinja2.meta import find_undeclared_variables

    if rendered_value not in _template_ast_cache:
        _template_ast_cache[rendered_value] = _j2_env.parse(rendered_value)
    ast = _template_ast_cache[rendered_value]

    return find_undeclared_variables(ast)



import difflib
from pathlib import Path

from openhasp_config_manager import util
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
from openhasp_config_manager.ui.util import print_diff_to_console, info, warn


class ConfigUploader:

    def __init__(self, output_root: Path, openhasp_client: OpenHaspClient):
        self._api_client = openhasp_client
        self._output_root = output_root
        self._cache_dir = Path(self._output_root, ".cache")

    def upload(self, device: Device, purge: bool = False, print_diff: bool = False) -> bool:
        """
        Uploads configuration files and config properties to a device.
        :param device: The device to upload to.
        :param purge: If True, removes files from the device, which are not present in the generated output.
        :param print_diff: If true, a diff will be printed to the console for each file that has changed.
        :return: True if any files have changed, false otherwise.
        """
        result = False

        if purge:
            result |= self.cleanup_device(device)

        result |= self._upload_files(device, print_diff)
        result |= self._update_config(device)
        return result

    def _upload_files(self, device: Device, print_diff: bool) -> bool:
        existing_files = self._api_client.get_files()

        result = False

        for file in device.output_dir.iterdir():
            info(f"Preparing '{file.name}' for upload...")

            if file.suffix in [".cmd", ".jsonl"]:
                result |= self._upload_text_file(device, print_diff, file, existing_files)
            else:
                result |= self._upload_binary_file(device, print_diff, file, existing_files)

        return result

    def _upload_text_file(self, device, print_diff, file, existing_files) -> bool:
        result = False

        content = file.read_text()
        if len(content) <= 0:
            warn(f"File is empty, skipping upload: {file}")
            return result

        # check if the checksum of the file has changed on the device
        file_content_on_device = ""
        if file.name in existing_files:
            file_content_on_device = self._api_client.get_file_content(file.name).decode("utf-8")
            device_file_content_checksum = util.calculate_checksum(file_content_on_device.encode("utf-8"))
        else:
            device_file_content_checksum = None

        new_checksum = self._check_if_checksum_will_change(
            file=file,
            original_checksum=device_file_content_checksum,
            new_content=content.encode("utf-8")
        )

        if new_checksum is not None:
            result = True
            if print_diff:
                diff_output = self._calculate_diff(
                    file_name=file.name,
                    string1=file_content_on_device,
                    string2=content
                )
                print_diff_to_console(diff_output)
            try:
                self._api_client.upload_file(file.name, content.encode("utf-8"))
                checksum_file = self._get_checksum_file(file)
                checksum_file.parent.mkdir(parents=True, exist_ok=True)
                checksum_file.write_text(new_checksum)
            except Exception as ex:
                raise Exception(f"Error uploading file '{file.name}' to '{device.name}': {ex}")
        else:
            info(f"Skipping {file} because it hasn't changed.")

        return result

    def _upload_binary_file(self, device, print_diff, file, existing_files) -> bool:
        result = False

        content = file.read_bytes()
        if len(content) <= 0:
            warn(f"File is empty, skipping upload: {file}")
            return result

        # check if the checksum of the file has changed on the device
        file_content_on_device = b""
        if file.name in existing_files:
            file_content_on_device: bytes = self._api_client.get_file_content(file.name)
            device_file_content_checksum = util.calculate_checksum(file_content_on_device)
        else:
            device_file_content_checksum = None

        new_checksum = self._check_if_checksum_will_change(
            file=file,
            original_checksum=device_file_content_checksum,
            new_content=content
        )

        if new_checksum is not None:
            result = True
            if print_diff:
                info(f"Binary file '{file.name}' has changed ({device_file_content_checksum} -> {new_checksum})")
            try:
                self._api_client.upload_file(file.name, content)
                checksum_file = self._get_checksum_file(file)
                checksum_file.parent.mkdir(parents=True, exist_ok=True)
                checksum_file.write_text(new_checksum)
            except Exception as ex:
                raise Exception(f"Error uploading file '{file.name}' to '{device.name}': {ex}")
        else:
            info(f"Skipping {file} because it hasn't changed.")

        return result

    def cleanup_device(self, device: Device) -> bool:
        """
        Delete files from the device, which are not present in the currently generated output
        :param device: the target device
        :return: True if any files have been deleted, false otherwise
        """
        result = False

        file_names = ["config.json"]
        for file in device.output_dir.iterdir():
            file_names.append(file.name)

        # cleanup files which are on the device, but not present in the generated output
        files_on_device = self._api_client.get_files()
        for f in files_on_device:
            if f not in file_names:
                result = True
                info(f"Deleting file '{f}' from device '{device.name}'")
                self._api_client.delete_file(f)
        return result

    def _check_if_checksum_will_change(self, file: Path, original_checksum: str, new_content: bytes) -> str | None:
        """
        Checks if the checksum for the given file has changed since it was last uploaded.
        :param file: the path of the file to check
        :param original_checksum: expected checksum of the original content
        :param new_content: the content of the new file
        :return: new checksum if the checksum changed, None otherwise
        """
        changed = False

        new_content_checksum = util.calculate_checksum(new_content)
        checksum_file = self._get_checksum_file(file)
        if not checksum_file.exists():
            changed = True
        else:
            old_hash = checksum_file.read_text().strip()
            if old_hash != original_checksum or old_hash != new_content_checksum:
                changed = True

        if changed:
            return new_content_checksum
        else:
            return None

    def _get_checksum_file(self, file: Path) -> Path:
        return Path(
            self._cache_dir,
            *file.relative_to(self._output_root).parts[:-1],
            file.name + ".md5"
        )

    def _update_config(self, device: Device):
        config_has_changed = False

        current_mqtt_config = self._api_client.get_mqtt_config()
        if current_mqtt_config == device.config.mqtt:
            info("MQTT config has not changed")
        else:
            info("Updating MQTT config...")
            self._api_client.set_mqtt_config(device.config.mqtt)
            config_has_changed = True

        current_http_config = self._api_client.get_http_config()
        if current_http_config == device.config.http:
            info("HTTP config has not changed")
        else:
            info("Updating HTTP config...")
            self._api_client.set_http_config(device.config.http)
            config_has_changed = True

        current_gui_config = self._api_client.get_gui_config()
        if current_gui_config == device.config.gui:
            info("GUI config has not changed")
        else:
            info("Updating GUI config...")
            self._api_client.set_gui_config(device.config.gui)
            config_has_changed = True

        return config_has_changed

    @staticmethod
    def _calculate_diff(file_name: str, string1: str, string2: str) -> str:
        diff = difflib.unified_diff(
            string1.splitlines(),
            string2.splitlines(),
            lineterm="\n",
            fromfile=f"${file_name}",
            tofile=f"${file_name}"
        )

        return "\n".join(list(diff))



from dataclasses import dataclass

from openhasp_config_manager.openhasp_client.model.configuration.debug_config import DebugConfig
from openhasp_config_manager.openhasp_client.model.configuration.gui_config import GuiConfig
from openhasp_config_manager.openhasp_client.model.configuration.hasp_config import HaspConfig
from openhasp_config_manager.openhasp_client.model.configuration.http_config import HttpConfig
from openhasp_config_manager.openhasp_client.model.configuration.mqtt_config import MqttConfig
from openhasp_config_manager.openhasp_client.model.configuration.telnet_config import TelnetConfig
from openhasp_config_manager.openhasp_client.model.configuration.wifi_config import WifiConfig
from openhasp_config_manager.openhasp_client.model.openhasp_config_manager_config import OpenhaspConfigManagerConfig


@dataclass
class Config:
    openhasp_config_manager: OpenhaspConfigManagerConfig
    wifi: WifiConfig
    mqtt: MqttConfig
    http: HttpConfig
    gui: GuiConfig
    hasp: HaspConfig
    debug: DebugConfig
    telnet: TelnetConfig



import asyncio
from pathlib import Path

from openhasp_config_manager.cli.common import _create_config_manager, _analyze_and_filter
from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
from openhasp_config_manager.ui.util import success, error, info


async def c_listen(config_dir: Path, device: str, path: str):
    try:
        config_manager = _create_config_manager(config_dir, Path("./nonexistent"))
        filtered_devices, ignored_devices = _analyze_and_filter(config_manager=config_manager, device_filter=device)
        if len(filtered_devices) <= 0:
            raise Exception(f"No device matches the filter: {device}")
        info(f"Listening to '.../{path}' on devices: {', '.join(map(lambda x: x.name, filtered_devices))}")
        for device in filtered_devices:
            client = OpenHaspClient(device)

            async def on_message(topic: str, payload: bytes):
                info(f"{topic}: {payload.decode('utf-8')}")

            await client.listen_event(path, on_message)

        asyncio.get_event_loop().run_forever()
        success("Done!")
    except Exception as ex:
        error(str(ex))



from typing import Dict


def calculate_checksum(content: bytes) -> str:
    import hashlib
    hash_value = hashlib.md5(content).hexdigest()
    return hash_value


def contains_nested_dict_key(d: dict, key: str) -> bool:
    """
    Recursively checks if any key in a nested dictionary structure is equal to the given string.
    :param d: the dictionary to check
    :param key: the key to search for
    :return: True if the string is found in any key, False otherwise
    """
    stack = [d]
    while stack:
        curr = stack.pop()
        for k, v in curr.items():
            if k == key:
                return True
            if isinstance(v, dict):
                stack.append(v)
    return False


def merge_dict_recursive(d1: Dict, d2: Dict) -> Dict:
    """
    Merges two dictionaries, preserving even the key/value pairs of nested dictionaries.

    This differs from the built-in "|" operator of python in that the built-in version
    will not preserve a key of a nested dict in d1 if the same dict is present in d2 but
    without the given key.

    :param d1: the first dictionary
    :param d2: the second dictionary
    :return: the merged dictionary
    """
    from copy import deepcopy

    result = deepcopy(d1)
    for d2_key, d2_value in d2.items():
        if d2_key in d1:
            d1_value = d1[d2_key]
            if isinstance(d1_value, dict):
                if not isinstance(d2_value, dict):
                    raise AssertionError(f"Incompatible types for merging dict, cannot merge {d1_value} and {d2_value}")

                result[d2_key] = merge_dict_recursive(d1_value, d2_value)
            else:
                result[d2_key] = deepcopy(d2_value)
        else:
            result[d2_key] = deepcopy(d2_value)

    return result



from typing import List

from PyQt6 import QtCore
from PyQt6.QtWidgets import QWidget, QPushButton, QHBoxLayout

from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.ui.qt.util import clear_layout


class DeviceListWidget(QWidget):
    deviceSelected = QtCore.pyqtSignal(Device)

    def __init__(self, devices):
        super().__init__()
        self.devices = devices
        self.layout = QHBoxLayout(self)
        self.create_entries()

    def set_devices(self, devices: List[Device]):
        self.devices = devices
        self.clear_entries()
        self.create_entries()

    def clear_entries(self):
        clear_layout(self.layout)

    def create_entries(self):
        for device in sorted(self.devices, key=lambda d: d.name):
            button = QPushButton(device.name)
            button.clicked.connect(lambda _, d=device: self.on_device_label_clicked(d))
            self.layout.addWidget(button)

    def on_device_label_clicked(self, device):
        print(f"Clicked on device: {device.name}")
        self.deviceSelected.emit(device)



from dataclasses import dataclass
from pathlib import Path
from typing import List

from openhasp_config_manager.openhasp_client.model.component import JsonlComponent, CmdComponent, \
    ImageComponent, FontComponent
from openhasp_config_manager.openhasp_client.model.configuration.config import Config


@dataclass
class Device:
    name: str
    path: Path
    config: Config
    jsonl: List[JsonlComponent]
    cmd: List[CmdComponent]
    images: List[ImageComponent]
    fonts: List[FontComponent]
    output_dir: Path



from pathlib import Path

from openhasp_config_manager.cli.common import _create_config_manager, _analyze_and_filter
from openhasp_config_manager.ui.util import error, success, info


async def c_screenshot(config_dir: Path, device: str, output: Path):
    try:
        config_manager = _create_config_manager(config_dir, Path("./nonexistent"))
        filtered_devices, ignored_devices = _analyze_and_filter(config_manager=config_manager, device_filter=device)

        if len(filtered_devices) <= 0:
            raise Exception(f"No device matches the filter: {device}")

        for device in filtered_devices:
            from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
            from openhasp_config_manager.uploader import ConfigUploader

            client = OpenHaspClient(device)
            try:
                info(f"Taking screenshot of device '{device.name}'...")
                screenshot = client.take_screenshot()
                image_file_path = Path(output, f"{device.name}.bmp")
                image_file_path.write_bytes(screenshot)
            except Exception as ex:
                raise Exception(f"Error taking screenshot of device '{device.name}': {ex}")

        success("Done!")
    except Exception as ex:
        error(str(ex))



from pathlib import Path
from typing import Dict, Any

import yaml
from yaml import Loader

from openhasp_config_manager.util import contains_nested_dict_key, merge_dict_recursive


class VariableManager:
    """
    Used to manage variable definitions found in YAML files inside the configuration
    directory.
    """

    def __init__(self, cfg_root: Path):
        self._cfg_root: Path = Path(cfg_root)
        self._path_vars: Dict[str, Dict] = {}

    def read(self):
        """
        Reads all variable definitions from the configuration directory.
        """
        self._path_vars = self._read(self._cfg_root)

    def add_var(self, key: str, value: any, path: Path = None):
        """
        Registers a variable to a path
        :param key: the variable key
        :param value: the variable value
        :param path: the path this variable should apply to
        """
        self.add_vars({key: value}, path)

    def add_vars(self, vars: Dict[str, Any], path: Path = None):
        """
        Registers a set of variables to a path
        :param vars: the variables
        :param path: the path the variables should apply to
        """
        if path is None:
            path = self._cfg_root

        relative_path = Path(self._cfg_root, path.relative_to(self._cfg_root))
        if relative_path.is_file():
            relative_path = relative_path.parent
        relative_path_str = str(relative_path)

        if relative_path_str not in self._path_vars.keys():
            self._path_vars[relative_path_str] = {}
        current_vars = self._path_vars.get(relative_path_str, {})
        combined = merge_dict_recursive(self._path_vars[relative_path_str], current_vars)
        combined = merge_dict_recursive(combined, vars)
        self._path_vars[relative_path_str] = combined

    def get_vars(self, path: Path) -> [str, Any]:
        """
        Returns the variable definitions and values for a given path.

        The resulting map will be a combination of all variables available within each
        of the directories along the given path. Variable definitions found in a lower (longer) path
        will override the ones that might be present higher up the path.

        :param path: the path context to use for variable evaluation
        :return: a map of "variable name" -> "variable value given the path context"
        """
        path_vars = self._get_vars_for_path(path)
        return path_vars

    def _get_vars_for_path(self, path: Path) -> Dict[str, Any]:
        """
        Returns the variable definitions and values for a given path.

        The resulting map will be a combination of all variables available within each
        of the directories along the given path. Variable definitions found in a lower (longer) path
        will override the ones that might be present higher up the path.

        :param path: the path context to use for variable evaluation
        :return: a map of "variable name" -> "variable value given the path context"
        """
        result = {}

        toplevel_path = self._cfg_root
        relative_path = Path(toplevel_path, path.relative_to(toplevel_path))
        if relative_path.is_file():
            relative_path = relative_path.parent

        current_path = None
        relative_paths = relative_path.parts
        for subfolder in relative_paths:
            if current_path is None:
                current_path = Path(subfolder)
            else:
                current_path = Path(current_path, subfolder)
            current_path_str = str(current_path)
            result = merge_dict_recursive(result, self._path_vars.get(current_path_str, {}))

        return result

    def _read(self, path: Path) -> Dict[str, Dict]:
        """
        Reads all "*.yaml" variable definition files in the given path
        :return: a map of "path -> dict of variables"
        """
        result = {}
        for toplevel_path in [path]:
            for p in list(toplevel_path.glob('**/**')):
                if not p.is_dir():
                    continue

                if p.name.startswith("."):
                    continue

                path_vars = self._create_vars_dict_for_path(p)

                if contains_nested_dict_key(path_vars, "items"):
                    # TODO: to avoid this, variables could be accessed by only exposing them
                    #  to jinja2 templates via a custom function like f.ex. "vars('my.key.items.a')".
                    #  This may be cumbersome to use though...
                    raise AssertionError(
                        "Variables contain key 'items' which conflics with the built-in function of jinja2. Please choose a different name.")

                path_str = str(p)
                if path_str not in result:
                    result[path_str] = {}
                result[path_str] = merge_dict_recursive(result[path_str], self._create_vars_dict_for_path(p))

        return result

    def _create_vars_dict_for_path(self, path: Path) -> Dict[str, Dict]:
        """
        Creates a dictionary containing all the variables for a given path by reading
        the yaml files in this path. This does _not_ take the path hierarchy
        into account. Only variables in the exact path will be returned in the result.
        :param path: the path to use as a context
        :return: a variable dictionary
        """
        result = {}
        path_var_files = list(path.glob("*.yaml"))
        for file in path_var_files:
            if not file.is_file():
                continue

            data = self._load_var_file(file)

            if data is None:
                # file is empty
                continue
            else:
                result = merge_dict_recursive(result, data)

        return result

    @staticmethod
    def _load_var_file(file: Path) -> Dict:
        content = file.read_text()
        return yaml.load(content, Loader=Loader)



from PyQt6.QtWidgets import QWidget, QMainWindow, QVBoxLayout

from openhasp_config_manager.manager import ConfigManager
from openhasp_config_manager.openhasp_client.model.component import CmdComponent
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.ui.qt.device_list import DeviceListWidget
from openhasp_config_manager.ui.qt.file_browser import FileBrowserWidget
from openhasp_config_manager.ui.qt.page_layout_editor import PageLayoutEditorWidget, OpenHaspDevicePagesData


class MainWindow(QMainWindow):
    def __init__(self, config_manager: ConfigManager):
        super().__init__()
        self.config_manager = config_manager

        self.setWindowTitle("openhasp-config-manager")

        self.create_basic_layout()
        self.load_plates()

    def create_basic_layout(self):
        self.container = QWidget()
        self.layout = QVBoxLayout()
        self.container.setLayout(self.layout)

        self.device_list_widget = DeviceListWidget(devices=[])
        self.device_list_widget.deviceSelected.connect(self.on_device_selected)
        self.layout.addWidget(self.device_list_widget)

        self.file_browser_widget = FileBrowserWidget(self.config_manager.cfg_root)
        self.layout.addWidget(self.file_browser_widget)

        self.page_layout_editor_widget = PageLayoutEditorWidget(
            config_manager=self.config_manager
        )
        self.layout.addWidget(self.page_layout_editor_widget)

        self.setCentralWidget(self.container)

    def load_plates(self):
        self.devices = self.config_manager.analyze()
        self.device_list_widget.set_devices(self.devices)

    def on_device_selected(self, device: Device):
        self.select_device(device)

    def select_device(self, device):
        self.device = device

        # setup sample page layout editor for testing
        # device_processor = self.config_manager.create_device_processor(device)
        # device_validator = self.config_manager.create_device_validator(device)
        self.relevant_components = self.config_manager.find_relevant_components(device)
        boot_cmd_component: CmdComponent = next(
            filter(lambda x: x.name == "boot.cmd", self.relevant_components), None)
        self.select_cmd_component(device, boot_cmd_component)

    def select_cmd_component(self, device: Device, cmd_component: CmdComponent):
        """
        Selects the given cmd component in the file browser widget.
        :param device: The device to select the cmd component for.
        :param cmd_component: The cmd component to select.
        """
        ordered_device_jsonl_components = self.config_manager.determine_device_jsonl_component_order_for_cmd(
            device=device,
            cmd_component=cmd_component,
        )

        self.page_layout_editor_widget.set_data(
            OpenHaspDevicePagesData(
                device=self.device,
                name=cmd_component.name,
                jsonl_components=ordered_device_jsonl_components
            )
        )



from pathlib import Path
from typing import Tuple, List

from openhasp_config_manager.manager import ConfigManager
from openhasp_config_manager.openhasp_client.model.device import Device
from openhasp_config_manager.processing.variables import VariableManager
from openhasp_config_manager.ui.util import info


async def _generate(config_manager: ConfigManager, device: Device):
    info(f"Generating output for '{device.name}'...")
    try:
        config_manager.process(device)
    except Exception as ex:
        raise Exception(f"Error generating output for {device.name}: {ex.__class__.__name__} {ex}")


async def _upload(device: Device, output_dir: Path, purge: bool, show_diff: bool):
    from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient
    from openhasp_config_manager.uploader import ConfigUploader

    client = OpenHaspClient(device)
    uploader = ConfigUploader(output_dir, client)

    info(f"Uploading files to device '{device.name}'...")
    return uploader.upload(device, purge, show_diff)


async def _deploy(config_manager: ConfigManager, device: Device, output_dir: Path, purge: bool, show_diff: bool):
    await _generate(config_manager, device)
    changed = await _upload(device, output_dir, purge, show_diff)
    # _cmd(config_dir, device="touch_down_1", command="reboot", payload="")
    # _reload(config_dir, device)
    if changed:
        info(f"Rebooting {device.name} to apply changes")
        await _reboot(device)
    else:
        info(f"No changes detected for {device.name}, device is already up-to-date")


async def _reload(device: Device):
    from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient

    client = OpenHaspClient(device)
    await client.command("clearpage", "all")
    await client.command("run", "L:/boot.cmd")


async def _reboot(device: Device):
    from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient

    client = OpenHaspClient(device)
    info(f"Rebooting {device.name}...")
    client.reboot()


async def _cmd(device: Device, command: str, payload: str):
    from openhasp_config_manager.openhasp_client.openhasp import OpenHaspClient

    client = OpenHaspClient(device)
    info(f"Sending command {command} to {device.name}...")
    await client.command(command, payload)


def _create_config_manager(config_dir, output_dir) -> ConfigManager:
    variable_manager = VariableManager(cfg_root=config_dir)
    config_manager = ConfigManager(
        cfg_root=config_dir,
        output_root=output_dir,
        variable_manager=variable_manager
    )
    return config_manager


def _analyze_and_filter(config_manager: ConfigManager, device_filter: str or None) -> Tuple[List[Device], List[Device]]:
    """

    :param config_manager:
    :param device_filter:
    :return: (matching_devices, ignored_devices)
    """
    info(f"Analyzing files in '{config_manager.cfg_root}'...")
    devices = config_manager.analyze()
    devices = list(sorted(devices, key=lambda x: x.name))
    device_names = list(map(lambda x: x.name, devices))
    info(f"Found devices: {', '.join(device_names)}")

    filtered_devices = []
    ignored_devices = []
    if device_filter is not None:
        filtered_devices = list(filter(lambda x: x.name == device_filter, devices))
        ignored_devices = list(filter(lambda x: x not in filtered_devices, devices))
    else:
        filtered_devices = devices

    return filtered_devices, ignored_devices
